#include <QApplication>
#include <QDesktopWidget>
#include <QAction>
#include <QFileDialog>
#include "cinespace.h"
#include "espace/timedialog.h"
#include "espace/confdialog.h"
#include "espace/starswidget.h"
#include "espace/TerreGLWidget.h"
#include "espace/univglwidget.h"
#include "espace/terre.h"
#include "espace/lune.h"
#include "espace/etoile.h"
#include "espace/planets.h"
#include "../mathutil/sp3.h"
#include "../utils/cjulien/julien.h"
#include <stdlib.h>  // pour system()
#include <math.h>
#include <QResource>
#include <QMessageBox>
#include <QString>
#include <QFile>
#include <QDebug>

// utilitaire copy fichier ressource pour acces depuis les vieilles routines en C
bool cpyRes(const QString &res,const char *nomfich)
{
    QResource *qr = new QResource(res);
    QFile qftempo(nomfich);
    if (!qftempo.open(QIODevice::WriteOnly))
    {
        QMessageBox::warning(0,"ERREUR OPEN","Echec ouverture tempo.txt");
        return false;
    }
    int n = qr->size();
    qftempo.write((const char *) qr->data(), n) ; qftempo.close();
    return true;
}

// Notre application, dérivant de Cinesim qui offre la fenêtre qui gère l'avancement du temps
CineSpace::CineSpace(QWidget *parent) : CineSim(parent)
{
    setWindowTitle(QObject::tr("CineSpace"));
    // On place cette fenetre en haut de l'ecran
    int h = minimumSizeHint().width() - myControl->minimumSizeHint().width();
    setFixedSize(QApplication::desktop()->width() - h, minimumSizeHint().height());
    move(0, 0); // Coin haut à gauche

    // On deconnecte Cinesim du signal isTime du CommandWidget (la connexion y est faite juste pour demo, si utilisé seul)
    disconnect(myControl, SIGNAL(isTime(int)), this, SLOT(timeOutSlot(int)));
    // On Affecte une fonction au bouton d'enregistrement
    connect(myControl->recButton,SIGNAL(clicked()),this,SLOT(recordSlot()));
    // Enregistrement
    recdown = false;
    // Timer de clignotement bouton
    clignoTimer = new QTimer(this);
    clignoTimer->stop();
    repEarth = new char[128];
    repStars = new char[128];
    fileEarth = new char[128];
    fileStars = new char[128];
    modeRec = RIEN;
    connect(clignoTimer, SIGNAL(timeout()), this, SLOT(clignoRecordSlot()));

    // On initialise la gestion du temps
    precision = 1.e-4;
    dtime_ms = 40; // intervalle de temps entre 2 affichage : 40 ms, pour 25 images par secondes
    accel = 7200.; // 2h temps simulé par seconde réelle
    dtsim = accel*dtime_ms*0.001; // intervalle de temps simu entre deux affichages
    // Instants début et fin quelconques
    time = tdeb = 0.25*86400.; // 6h après J2000 (1/1/2000 à 12h + 6h -> 1/1/2000 à 18h)
    tfin = 86400*100.; // 100 jours après J2000 (cad le 10/4/2000 à 12h)
    // Les modificateurs de l'ecoulement du temps
    periode = 86400.; // Periodicité de surveillance d'un évènement
    creneau = 86400.; // Duree d'observation dans la période
    scenar = NOSCENE; // Modification éventuelle du début de la période pour se caler à un scenario
    // (lever ou coucher du Soleil par ex.)
    decalage = 0.;    // Décalage du début d'observation par rapport au temps calculé
    satcou = 0;       // Satellite courant (si des satellites sont chargés)

    // Position de l'observateur (modifiable par le ConfDialog) lati = 45., longi = 0.;
    Terre::setLongiLati(0., M_PI/4.) ;

    // La position des planètes sera calculée dans le reférentiel de la date, méthode sphérique.
    Planets::config(false, true);

    // Conteneurs vides pour les TLE et les Satellites
    all_tle = new std::vector<TwoLE *>();
    all_sats = new std::vector<SatGeo *>();

    myControl->setTimer(dtime_ms);
    synchroTimeDialogSlot(); // Calcul numero du pas d'horloge courant et du nombre de pas d'horloge total
    // On connecte l'avancement du temps à la mise à jour des scenes a dessiner
    connect(myControl, SIGNAL(isTime(int)), this, SLOT(majSceneSlot(int)));

    // On cree le TimeDialog qui permet de modifier les paramètres temporels
    timeDialog = new TimeDialog(this, this); timeDialog->setModal(false);
    connect(timeDialog, SIGNAL(doSynchro()), this, SLOT(synchroTimeDialogSlot()));
    // On ajoute un item au menu pour l'afficher
    QAction *timeDialogAction = new QAction(tr("&Ordonnancement"), this);
    connect(timeDialogAction, SIGNAL(triggered()), this, SLOT(showTimeDialogSlot()));
    addAction(timeDialogAction);

    // On cree le ConfDialog qui permet de configurer l'observateur et ce qu'il désire voir
    confDialog = new ConfDialog(this, this); confDialog->setModal(false);
    // On met le satellite bidon dans la liste des satellites
    SatGeo *sat = new SatGeo();
    if(cpyRes(":/bidon", "tempo.txt") && sat->set("tempo.txt"))
    {
        all_sats->push_back(sat);
        confDialog->makeListViewSatsSel();
    }
    // Modif des permissions pour pouvoir effacer le fichier
    QFile::setPermissions("tempo.txt", (QFileDevice::Permission) 0x6666);
    QFile::remove("tempo.txt");

    confDialog->makeListViewSatsSel();
    // On ajoute un item au menu pour l'afficher
    QAction *confDialogAction = new QAction(tr("&Configuration"), this);
    connect(confDialogAction, SIGNAL(triggered()), this, SLOT(showConfDialogSlot()));
    addAction(confDialogAction);

    // On ajoute un item au menu pour sauver le scenario
    QAction *saveSceneAction = new QAction(tr("&Ecriture Scenario"), this);
    connect(saveSceneAction, SIGNAL(triggered()), this, SLOT(saveSceneSlot()));
    addAction(saveSceneAction);
    // On ajoute un item au menu pour lecture scenario
    QAction *loadSceneAction = new QAction(tr("&Lecture Scenario"), this);
    connect(loadSceneAction, SIGNAL(triggered()), this, SLOT(loadSceneSlot()));
    addAction(loadSceneAction);

    labelHelp = new QLabel(
                tr(
                    "<h4>Interface de commande :</h4>"
                    "<ul>"
                    "<li>Bouton REC : Pour faire un enregistrement des images en vue de faire une vidéo.<br/>"
                    "Une seule image est enregistrée à chaque pas d'avance de la simulation.</li>"
                    "<li>Bouton PLAY/PAUSE : Pour faire avancer ou stopper la simulation.</li>."
                    "<li>Bouton UN_PAS_DIR/Un_PAS_INV : pour faire avancer/reculer la simulation d'un seul pas.</li>"
                    "<li>Bouton INVERSER : Pour inverser le sens de la simulation.</li>"
                    "<li>BOUTON INIT : Pour revenir au début</li>"
                    "<li>Le numéro du pas encours et le nombre de pas total est affiché entre le bouton INIT et le curseur.<br/>"
                    "Le nombre de pas total dépend de l'ordonnancement choisi (voir menu Ordonnancement).<br/>"
                    "Le curseur permet de commander le numéro des pas dans les deux sens.</li>"
                    "</ul>"
                    "<h4>Menu contextuel :</h4>"
                    "<ul>"
                    "<li>Ordonnancement :<br/>Affiche une boite de dialogue permettant de modifier les dates"
                    " et tranches d'observation.</li>"
                    "<li>Configuration :<br/>Pour ajouter des satellites et modifier la position de l'observateur.</li>"
                    "<li>Ecriture scenario :<br/>Pour enregistrer les paramètres de configuration et ordonnancement actuels.</li>"
                    "<li>Lecture scenario :<br/>Pour acquérir les paramètres d'un scenario pré-enregistré.</li>"
                    "<li>Affichage ciel étoilé :<br/>Affiche une vue du ciel étoilé depuis un point d'observation sur Terre.</li>"
                    "<li>Affichage Terre :<br/>Affiche la Terre ou la Lune selon divers point de vue.</li>"
                    "<li>Help :<br/>Affiche cette aide.</li>"
                    "</ul>"
                    ));
    labelHelp->setWindowTitle("Aide CineSpace");

    // On crée l'affichage des étoiles
//    LectureEtoilesEtConstellations(&tabStars, &tabConstellations);
    tabStars = Etoile::init();
    if (tabStars==NULL) exit(1);
    tabConstellations = Constellation::init();
    if(tabConstellations==NULL) exit(1);
    winStars = new StarsWidget(720,500, tabStars, tabConstellations, this);
    // On ajoute un item au menu pour l'afficher
    QAction *starsWidgetAction = new QAction(tr("&Affichage ciel étoilé"), this);
    connect(starsWidgetAction, SIGNAL(triggered()), this, SLOT(showStarsSlot()));
    addAction(starsWidgetAction);

    // On cree l'affichage Terre
    winTerre = new TerreGLWidget(this);
    // On ajoute un item au menu pour l'afficher
    QAction *showTerreAction = new QAction(tr("&Affichage Terre"), this);
    connect(showTerreAction, SIGNAL(triggered()), this, SLOT(showTerreSlot()));
    addAction(showTerreAction);

    // On cree l'affichage Univers
    winUniv = new UnivGLWidget(this);
    // On ajoute un item au menu pour l'afficher
    QAction *showUnivAction = new QAction(tr("&Affichage Système solaire"), this);
    connect(showUnivAction, SIGNAL(triggered()), this, SLOT(showUnivSlot()));
    addAction(showUnivAction);

    // item "Help"
    QAction *actionHelp = new QAction(tr("&Help"), this);
    connect(actionHelp, SIGNAL(triggered()), this, SLOT(afficheHelpSlot()));
    addAction(actionHelp);

    // Mise à jour de la scene initiale
    majSceneSlot(nech);
}

// Reception clic demande d'affichage du Terre, Stars, Dialog d'ordonnancement
void CineSpace::showTerreSlot(){winTerre->show();}
void CineSpace::showUnivSlot(){winUniv->show();}
void CineSpace::showStarsSlot(){winStars->show();}
void CineSpace::showTimeDialogSlot() {timeDialog->show();}
void CineSpace::showConfDialogSlot() {confDialog->show();}

// Clignotement bouton enregistrement
void CineSpace::clignoRecordSlot()
{
    recdown = !recdown;
    if (recdown) myControl->recButton->setStyleSheet("background-color : red");
    else myControl->recButton->setStyleSheet("");
}

// Gestion bouton Record
void CineSpace::recordSlot()
{
    if (modeRec != RIEN)
    { // Arret enregistrement
        clignoTimer->stop();
        recdown = false;
        myControl->recButton->setStyleSheet("");
        if(QMessageBox::Yes == QMessageBox::question(this,"Cinespace","Avez-vous ffmpeg accessible dans le path ?"))
        {
            char strExe[256];
            int result;
            if (modeRec == LES2 || modeRec == EARTH)
            {
                sprintf(strExe,"ffmpeg -f image2 -i %s/img%%06d.jpg -vcodec mpeg4 -b 3000k %s/terre.mp4",
                        repEarth, repEarth);
                result = system(strExe);
                if (result == 0)
                {
                    if(QMessageBox::Yes == QMessageBox::question(this,"Cinespace","Voulez-voir la video de la fenêtre Terre ?"))
                    {
                        sprintf(strExe,"ffplay %s/terre.mp4", repEarth);
                        system(strExe);

                    }
                }
                else QMessageBox::information(this, "CineSpace", "Il y a eu un problème avec ffmpeg");
            }
            if (modeRec == LES2 || modeRec == STARS)
            {
                sprintf(strExe,"ffmpeg -f image2 -i %s/img%%06d.jpg -vcodec mpeg4 -b 3000k %s/stars.mp4",
                        repStars, repStars);
                result = system(strExe);
                if (result == 0)
                {
                    if(QMessageBox::Yes == QMessageBox::question(this,"Cinespace","Voulez-voir la video de la fenêtre Ciel ?"))
                    {
                        sprintf(strExe,"ffplay %s/stars.mp4", repStars);
                        system(strExe);
                    }
                }
                else QMessageBox::information(this, "CineSpace", "Il y a eu un problème avec ffmpeg");
            }
        }
        else
        {
            QMessageBox::information(this,
                                     "CineSpace",
                                     "Pour faire une video avec les images enregistrées télechargez ffmpeg\n"
                                     "Exemple de commande pour faire une video avec ffmpeg :\n"
                                     "ffmpeg -f image2 -i img%06d.jpg -vcodec mpeg4 -b 3000k video.mp4");
        }
        modeRec = RIEN;
    }
    else // Pour le moment modeRec == RIEN
    {
        // ouverture fichier enregistrement
        ModePlay oldmp =  myControl->modePlay;
        // Si en Play, on met en Pause
        if (myControl->modePlay == PLAY) myControl->playpauseSlot();

        //Question si trajectoire, video, les 2
        QMessageBox mb(QMessageBox::Information,"QUESTION ?","Voulez-vous enregistrer");
        QPushButton *recStars = mb.addButton("Ciel étoilé", QMessageBox::ActionRole);
        QPushButton *recEarth = mb.addButton("Vue Terre", QMessageBox::ActionRole);
        QPushButton *recBoth = mb.addButton("Les 2", QMessageBox::ActionRole);
        mb.addButton(QMessageBox::Abort);
        mb.exec();
        QPushButton *res = (QPushButton *) mb.clickedButton();
        // Lecture nom fichier pour enregistrer le ciel étoilé
        if(res == recStars || res == recBoth)
        {
            QString repRec = QFileDialog::getExistingDirectory
                    (NULL, tr("Repertoire enregistrement ciel étoilé ?"), ".",
                     QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
            if(!repRec.isEmpty())
            {
                if (modeRec == RIEN ) modeRec = STARS;
                else modeRec = LES2;
                strcpy(repStars, repRec.toLocal8Bit().constData());
                strcpy(fileStars, repStars);
                strNumImgStars = fileStars + strlen(fileStars);
                numImgStars = 1;
            }
        }
        // Lecture nom repertoire pour enregistrer les images
        if(res == recEarth || res == recBoth)
        {
            QString repRec = QFileDialog::getExistingDirectory
                    (NULL, tr("Repertoire enregistrement scène spatiale ?"), ".",
                     QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
            if(!repRec.isEmpty())
            {
                if (modeRec == RIEN ) modeRec = EARTH;
                else modeRec = LES2;
                strcpy(repEarth, repRec.toLocal8Bit().constData());
                strcpy(fileEarth, repEarth);
                strNumImgEarth = fileEarth + strlen(fileEarth);
                numImgEarth = 1;
            }
        }
        if (oldmp == PLAY) myControl->playpauseSlot();
        if (modeRec == RIEN) return;
        clignoTimer->setInterval(250);
        clignoTimer->start();
    }
}

void CineSpace::closeEvent(QCloseEvent *event)
{
    if(timeDialog) timeDialog->close();
    if(confDialog) confDialog->close();
    if(winUniv->isVisible()) winUniv->close();
    if(winTerre->isVisible()) winTerre->close();
    if(winStars->isVisible()) winStars->close();
    labelHelp->close();
}

// Synchronisation des paramètres temporels suite modif operateur
// avec eventuellement modification de l'acceleration
void CineSpace::synchroTimeDialogSlot()
{
    // Le créneau ne peut être supérieur à la période
    if (creneau > periode) {creneau = periode; timeDialog->setCreneau(creneau);}
    memTranche = -1;
    deltaCreneau = 0;
    dtsim = accel*dtime_ms*0.001;
    // Si l'intervalle de temps simu entre 2 échantillons est plus grand que le créneau
    // on ralentit l'accélération pour ramener cet intervalle à la largeur du créneau
    if (dtsim > creneau)
    {
        dtsim = creneau;
        accel = dtsim/(dtime_ms*0.001); timeDialog->setAccel(accel);
    }

    // Recalcul du nombre d'échantillons sur la nouvelle durée et, par ailleurs,
    // on essaie de calculer un numéro d'échantillon qui préserve l'instant actuel de la simu
    // de manière à avoir une continuité de la simu. Si on n'y arrive pas, on se positionne au début.
    double duree = tfin - tdeb;     // Durée totale simulée
    int n = duree/periode;      // Nombre de tranches simulées
    double reste = duree - n*periode; // Temps simulé dans la tranche numéro n+1
    if (reste > creneau) reste = creneau;
    nechtot = (n*creneau + reste)/dtsim;  // Nombre d'échantillons dans l'intervalle simulé
    if (time < tdeb || time >= tfin) time = tdeb; // Correction temps courant si hors domaine
    duree = time - tdeb;
    n = duree/periode; // Nouveau numéro tranche en cours
    reste = duree - n*periode;
    if (reste > creneau) reste = creneau;
    nech = (n*creneau + reste)/dtsim; // numéro échantillon correspondant au temps courant
    myControl->setSlider(nech, nechtot);
}

/*-------------------------------------
                 * TRANSFORMATION     nech -> time
                 *-------------------------------------
                Réception des paramètres envoyés par le gestionnaire de l'avance de la simulation
                qui permet de calculer le temps courant simu :
                - en multipliant par dtsim (durée de l'intervalle de temps simulé entre 2 échantillons)
                - en ajoutant l'instant début
                ce qui donne le temps simulé non corrigé.
                Si creneau = période, il n'y a pas de correction du temps, sinon, la durée de la simulation
                est divisée en tranches de durées = période et dans chaque tranche on ne simule qu'une durée
                de temps égale à créneau, puis on passe instantanément à la tranche suivante. Dans une tranche
                (de durée période) ce temps corrigé avance normalement pendant une durée "créneau", puis saute
                instantanément au début de la tranche suivante.
                Par ailleurs, pour les scenarios LEVER et/ou COUCHER, on corrige les instants de début de
                tranche pour qu'ils coïncident avec le lever et/ou coucher du Soleil le plus proche de la
                valeur deltaCreneau (mémorisée car variable de classe).
                Enfin on ajoute "decalage" choisit par l'utilisateur (pour modifier le début de la tranche
                par rapport à l'instant du lever et/ou coucher.
                */
void CineSpace::majSceneSlot(int nech)
{
    int n;

    double time = nech*dtsim; // Temps simu relatif non corrigé
    n = time/creneau;     // Numéro de la tranche
    time += n*(periode-creneau) + tdeb; // // Temps simu relatif corrigé
    // Correction supplémentaires si début de tranche synchronisé avec lever ou coucher
    if (scenar != NOSCENE)
    {
        // Qd on change de tranche, on recale son début sur un lever ou coucher du Soleil
        if (n != memTranche)
        {
            memTranche = n;
            double tempo;
            tsunhzero tsun;
            double jours = time/86400.;
            double te_j = jours + jul2deltaET(jours + 2451545.);
            Terre::sunhzero(&tsun, jours, te_j, Terre::longobs, Terre::latobs, scenar-NOSCENE);
            if (scenar == LEVER) tempo = tsun.leverUT*86400.;
            if (scenar == COUCHER) tempo = tsun.coucherUT*86400.;
            deltaCreneau = tempo - time;
        }
        time += deltaCreneau;
    }
    time += decalage;
    update(time);
    if (winTerre->isVisible() && (modeRec == EARTH || modeRec == LES2))
    {
        sprintf(strNumImgEarth,"/img%06d.jpg",numImgEarth++);
        QImage image = winTerre->grabFrameBuffer();
        image.save(QString(fileEarth));
        if (numImgEarth >= 200000) recordSlot(); // pour arreter
    }
    if (winStars->isVisible() && (modeRec == STARS || modeRec == LES2))
    {
        sprintf(strNumImgStars,"/img%06d.jpg",numImgStars++);
        QImage image(winStars->size(), QImage::Format_ARGB32_Premultiplied);
        QPainter pt(&image);  winStars->paintScene(pt);
        image.save(QString(fileStars));
        if (numImgStars >= 200000) recordSlot(); // pour arreter
    }
}

void CineSpace::update(double t)
{
    // maj temps
    time = t;
    // maj affichage temps dans le timedialog
    if (timeDialog->isVisible()) timeDialog->updateCourant(time);
    double tu_j = time/86400.;
    double te_j = tu_j + jul2deltaET(tu_j + 2451545.);
    // maj orientation Terre et position Soleil
    Terre::posSoleil(te_j, precision);
    Terre::orientation(tu_j, precision);
    // maj position Lune et repère Lune
//    if(winStars->drawLune || winTerre->drawLune || winUniv->isVisible())
    {
        Lune::position(te_j, precision);
        Lune::phases();
        if (winTerre->drawLune || winUniv->isVisible())
            Lune::orientation(te_j, precision);
    }

    // maj position planètes
//    if(winStars->drawPlanetes || winUniv->isVisible())
        Planets::etat(te_j, precision);
    // maj position satellites
    if(winStars->drawSatellites || winTerre->drawSatellites)
        for(int k = 0; k < all_sats->size(); k++) all_sats->at(k)->etat(time, precision);
    // maj direction étoiles sans correction nutation, faite au niveau orientation Terre.
    Etoile::misajour(te_j, precision, false);
    // maj des tracés
    if (winStars->isVisible())
        winStars->update();
    if (winTerre->isVisible()) winTerre->updateGL();
    if (winUniv->isVisible()) winUniv->updateGL();
}

//-------------------------------------------
// LECTURE DU LOT DE PARAMETRES D'UN SCENARIO
//-------------------------------------------

static const char *const varnames[] =
{
    "tdeb", "tfin", "accel", "periode", "creneau", "longitude", "latitude", "aziVisee", "sitVisee",
    "alfamax", "magmin", "clarte", "flagHoriz", "flagChamp",
    "drawConstel", "drawStarName", "drawSgName", "drawConsName", "scenar", "decalage", "kmulSL",
    "drawLune", "drawPlanetes", "drawSatellites", "terrePDV", "precision",
    "gainDist, gainAngle", "aziScene", "sitScene","perspective", 0
};

static const int nbnames = sizeof(varnames)/sizeof(char *) - 1;


// Reception clic menu Ecriture Scenario
void CineSpace::saveSceneSlot()
{
    QString NomFich = QFileDialog::getSaveFileName(NULL, QObject::tr("Fichier Ecriture Scenario ?"),".");
    if (NomFich.isNull() || NomFich.isEmpty()) return;
    FILE *fdata = fopen(NomFich.toLatin1().constData(),"wt");
    if (fdata == NULL) return;

    for (int i = 0 ; i < nbnames; i++)
    {
        switch (i)
        {
        case 20 :
            fprintf(fdata,"%s = %lf #Coef. mult. taille Soleil et Lune [1, 10]\n", varnames[i], winStars->kmulSL);
            break;
        case 18 :
            fprintf(fdata,"%s = %d #Scenario : 0=aucun, 1=lever, 2=coucher\n", varnames[i], (int) scenar);
            break;
        case  19 :
            fprintf(fdata,"%s = %lf #Décalage temporel en secondes\n", varnames[i], decalage);
            break;
        case 0 :
        case 1 : // tdeb, tfin
            double val;
            if(i) val= tfin; else val = tdeb;
            val = val/86400. + Date_J2000_j;
            int date[6];
            jul2tdat(date, val);
            fprintf(fdata,"%s = %d %d %d %d %d %d #année mois jour heures minutes secondes\n", varnames[i],
                    date[0], date[1], date[2], date[3], date[4], date[5]);
            break;
        case  2 :
            fprintf(fdata,"%s = %lf #Accélération du temps si >0 et nb d'ech/ période si <0\n", varnames[i], accel);
            break;
        case  3 :
            fprintf(fdata,"%s = %lf #Période en jours de répétition de la simulation\n", varnames[i], periode/86400.);
            break;
        case  4 :
            fprintf(fdata,"%s = %lf #Durée en secondes de la simulation par période\n", varnames[i], creneau);
            break;
        case  5 :
            fprintf(fdata,"%s = %lf #Longitude de l'observateur en degrés décimaux [-180, +180]\n", varnames[i], Terre::longobs*180./M_PI);
            break;
        case  6 :
            fprintf(fdata,"%s = %lf #Latitude de l'observateur en degrés décimaux [-90, +90]\n", varnames[i], Terre::latobs*180./M_PI);
            break;
        case  7 :
            fprintf(fdata,"%s = %lf #Azimut par rapport au Sud de la visée en degrés décimaux [-180, +180]\n", varnames[i], winStars->aziVisee*180./M_PI);
            break;
        case  8 :
            fprintf(fdata,"%s = %lf #Site de la visée en degrés décimaux [0, 90]\n", varnames[i], winStars->sitVisee*180./M_PI);
            break;
        case  9 :
            fprintf(fdata,"%s = %lf #Champs de vue en degrés décimaux [0.01, 180]\n", varnames[i], winStars->alfamax*180./M_PI);
            break;
        case 10 :
            fprintf(fdata,"%s = %lf #Magnitude des étoiles observées [0, 8]\n", varnames[i], winStars->magmin);
            break;
        case 11 :
            fprintf(fdata,"%s = %d #Couleur du ciel : 0=noir, 1=blanc, 2=selon l'éclairement par le soleil\n", varnames[i], (int) winStars->clarte);
            break;
        case 12 :
            fprintf(fdata,"%s = %d #Etoiles cachées par le sol visibles si 0\n", varnames[i], winStars->flagHoriz ? 1 : 0);
            break;
        case 13 :
            fprintf(fdata,"%s = %d #Etoiles cachées par le champ de vue visibles si 0\n", varnames[i], winStars->flagChamp ? 1 : 0);
            break;
        case 14 :
            fprintf(fdata,"%s = %d #Constellations non tracées si 0\n", varnames[i], winStars->drawConstel ? 1 : 0);
            break;
        case 15 :
            fprintf(fdata,"%s = %d #Noms des principales étoiles non écrits si 0\n", varnames[i], winStars->drawStarName ? 1 : 0);
            break;
        case 16 :
            fprintf(fdata,"%s = %d #Noms courts de toutes les étoiles non écrits si 0\n", varnames[i], winStars->drawSgName ? 1 : 0);
            break;
        case 17 :
            fprintf(fdata,"%s = %d #Noms des constellations nons écrits si 0\n", varnames[i], winStars->drawConsName ? 1 : 0);
            break;
        case 21 :
            fprintf(fdata,"%s = %d #Lune non tracée si 0\n", varnames[i], winStars->drawLune || winTerre->drawLune ? 1 : 0);
            break;
        case 22 :
            fprintf(fdata,"%s = %d #Planètes non tracées si 0\n", varnames[i], winStars->drawPlanetes ? 1 : 0);
            break;
        case 23 :
            fprintf(fdata,"%s = %d #Satellites non tracés si 0\n", varnames[i], winStars->drawSatellites || winTerre->drawSatellites ? 1 : 0);
            break;
        case 24 :
            fprintf(fdata,"%s = %d #Point de vue fenêtre Terre\n", varnames[i], (int) winTerre->modePtVue);
            break;
        case 25 :
            fprintf(fdata,"%s = %lf #Précision générale\n", varnames[i], precision);
            break;
        case 26 :
            fprintf(fdata,"%s = %lf #Gain distance visu Terre\n", varnames[i], winTerre->gainDist);
            break;
        case 27 :
            fprintf(fdata,"%s = %lf #Gain angle visu Terre\n", varnames[i], winTerre->gainAngle);
            break;
        case 28 :
            fprintf(fdata,"%s = %lf #Azimut scene visu Terre\n", varnames[i], winTerre->aziScene);
            break;
        case 29 :
            fprintf(fdata,"%s = %lf #Sit. scene visu scene\n", varnames[i], winTerre->sitScene);
            break;
        case 30 :
            fprintf(fdata,"%s = %d #Flag perspective visu scene\n", varnames[i], winTerre->perspective);
            break;
        }
    }
    fclose(fdata);
}


// Enlève les blancs de tête et de queue (et les fin de ligne aussi). Renvoie un pointeur sur le début valide
// et met 0 après le dernier caractère non espace
static char *trimme(char *str)
{
    int first = 0, last = strlen(str);
    while (first < last && isspace(str[first])) first++;
    while (last > first && isspace(str[last-1]))last--;
    str[last]=0;
    return str+first;
}

// Reception clic menu Lecture Scenario
void CineSpace::loadSceneSlot()
{

    winStars->initConf(); // Au cas où certains paramètres ne seraient pas trouvés dans le scenario lu

    int flagNok = 0;
    char ligne[256], *str, *nom;
    double temps, val;
    int i, k;
    double val_tdeb=0., val_tfin=0., val_accel=0., val_periode=0., val_creneau=0., val_longitude, val_latitude, val_aziVisee, val_sitVisee,
            val_alfamax, val_magmin, val_decalage, val_kmulSL;
    int val_clarte, val_flagHoriz, val_flagChamp, val_drawConstel, val_drawStarName, val_drawSgName, val_drawConsName;
    int val_scenar, val_drawLune, val_drawPlanetes, val_drawSatellites, val_terrePDV, val_perspective;
    double val_precision, val_gainDist, val_gainAngle, val_aziScene, val_sitScene;

    // On prend les anciennes valeurs par défaut.
    val_tdeb = tdeb; val_tfin = tfin; val_accel = accel; val_periode = periode; val_creneau = creneau;
    val_longitude = Terre::longobs; val_latitude = Terre::latobs; val_aziVisee = winStars->aziVisee; val_sitVisee = winStars->sitVisee;
    val_alfamax = winStars->alfamax; val_magmin = winStars->magmin; val_clarte = winStars->clarte; val_kmulSL = winStars->kmulSL;
    val_flagHoriz = winStars->flagHoriz; val_flagChamp = winStars->flagChamp; val_drawConstel = winStars->drawConstel;
    val_drawStarName = winStars->drawStarName; val_drawSgName = winStars->drawSgName; val_drawConsName = winStars->drawConsName;
    val_scenar = scenar;
    val_decalage = decalage;
    val_drawLune = winStars->drawLune || winTerre->drawLune;
    val_drawPlanetes = winStars->drawPlanetes;
    val_drawSatellites = winStars->drawSatellites || winTerre->drawSatellites;
    val_terrePDV = (int) winTerre->modePtVue;
    val_precision = precision;
    val_gainDist = winTerre->gainDist;
    val_gainAngle = winTerre->gainAngle;
    val_aziScene = winTerre->aziScene;
    val_sitScene = winTerre->sitScene;
    val_perspective = winTerre->perspective;

    // on lit les nouvelles
    QString NomFich = QFileDialog::getOpenFileName (NULL,
                                                    QObject::tr("Fichier Lecture Scenario ?"),".");
    if (NomFich.isNull() || NomFich.isEmpty()) return;
    FILE *fdata = fopen(NomFich.toLatin1().constData(),"rt");
    if (fdata == NULL) return;
    while(!feof(fdata) && fgets(ligne, 255, fdata) != NULL)
    {
        str = strtok(ligne, "="); nom = trimme(str);
        for (i = 0 ; i < nbnames; i++)
        {
            if(strcmp(nom, varnames[i]) == 0)
            {
                switch (i)
                {
                case 0 :
                case 1 : // tdeb, tfin
                    int date[6];
                    for (k = 0; k < 6 ; k++)
                    {
                        str=strtok(NULL, " \t");
                        if (str == NULL) break;
                        date[k] = atoi(trimme(str));
                    }
                    if (k < 3) {flagNok = i+1; break;}
                    while(k < 6) date[k++] = 0;
                    temps = (tdat2jul(date) - Date_J2000_j)*86400.;
                    if(i) val_tfin = temps; else val_tdeb = temps;
                    break;
                case  2 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_accel = strtod(trimme(str), NULL);
                    break;
                case  3 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_periode = 86400.*strtod(trimme(str), NULL);
                    if (val_periode <= 0.) flagNok = i+1;
                    break;
                case  4 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_creneau = strtod(trimme(str), NULL);
                    if (val_creneau <= 0.) flagNok = i+1;
                    break;
                case  5 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val = strtod(trimme(str), NULL);
                    if (val < -180. || val > 180.) {flagNok = i+1; break;}
                    val_longitude = (M_PI/180.)*val;
                    break;
                case  6 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val = strtod(trimme(str), NULL);
                    if (val < -90. || val > 90.) {flagNok = i+1; break;}
                    val_latitude = (M_PI/180.)*val;
                    break;
                case  7 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val = strtod(trimme(str), NULL);
                    if (val < -180. || val > 180.) {flagNok = i+1; break;}
                    val_aziVisee = (M_PI/180.)*val;
                    break;
                case  8 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val = strtod(trimme(str), NULL);
                    if (val < 0. || val > 90.) {flagNok = i+1; break;}
                    val_sitVisee = (M_PI/180.)*val;
                    break;
                case  9 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val = strtod(trimme(str), NULL);
                    if (val < 0.01 || val > 180.) {flagNok = i+1; break;}
                    val_alfamax = (M_PI/180.)*val;
                    break;
                case 10 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val = strtod(trimme(str), NULL);
                    if (val < 0. || val > 8.) {flagNok = i+1; break;}
                    val_magmin = val;
                    break;
                case 11 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_clarte = atoi(trimme(str));
                    if (val_clarte < 0 || val_clarte > 2) flagNok = i+1;
                    break;
                case 12 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_flagHoriz = atoi(trimme(str)) ? true : false;
                    break;
                case 13 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_flagChamp = atoi(trimme(str)) ? true : false;
                    break;
                case 14 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_drawConstel = atoi(trimme(str)) ? true : false;
                    break;
                case 15 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_drawStarName = atoi(trimme(str)) ? true : false;
                    break;
                case 16 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_drawSgName = atoi(trimme(str)) ? true : false;
                    break;
                case 17 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_drawConsName = atoi(trimme(str)) ? true : false;
                    break;
                case 18 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_scenar = atoi(trimme(str));
                    if (val_scenar < 0 || val_scenar > 2) flagNok = i+1;
                    break;
                case  19 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_decalage = strtod(trimme(str), NULL);
                    break;
                case  20 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_kmulSL = strtod(trimme(str), NULL);
                    break;
                case 21 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_drawLune = atoi(trimme(str)) ? true : false;
                    break;
                case 22 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_drawPlanetes = atoi(trimme(str)) ? true : false;
                    break;
                case 23 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_drawPlanetes = atoi(trimme(str)) ? true : false;
                    break;
                case 24 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_terrePDV = atoi(trimme(str));
                    break;
                case 25 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_precision = strtod(trimme(str), NULL);
                    break;
                case 26 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_gainDist = strtod(trimme(str), NULL);
                    break;
                case 27 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_gainAngle = strtod(trimme(str), NULL);
                    break;
                case 28 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_aziScene = strtod(trimme(str), NULL);
                    break;
                case 29 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_sitScene = strtod(trimme(str), NULL);
                    break;
                case 30 :
                    str=strtok(NULL, " \t");
                    if (str == NULL) {flagNok = i+1; break;}
                    val_perspective = atoi(trimme(str));
                    break;

                }
                break;
            }
            if (flagNok) break;
        }
        if (flagNok) break;
    }
    fclose(fdata);
    if (flagNok)
    {
        qDebug() << "Erreur lecture variable " << varnames[i] << endl; // On conserve les anciennes valeurs
        return;
    }

    // On accepte les nouvelles valeurs
    tdeb = val_tdeb; tfin = val_tfin; accel = val_accel; periode = val_periode; creneau = val_creneau;
    winStars->aziVisee = val_aziVisee; winStars->sitVisee = val_sitVisee;
    Terre::setLongiLati(val_longitude, val_latitude); confDialog->initLongiLati(val_longitude, val_latitude);
    winStars->alfamax = val_alfamax; winStars->magmin = val_magmin; winStars->clarte = (Eclairage) val_clarte;
    winStars->flagHoriz = val_flagHoriz; winStars->flagChamp = val_flagChamp; winStars->drawConstel = val_drawConstel;
    winStars->drawStarName = val_drawStarName; winStars->drawSgName = val_drawSgName; winStars->drawConsName = val_drawConsName;
    scenar = (Scenario) val_scenar;
    decalage = val_decalage;
    winStars->kmulSL = val_kmulSL;
    winStars->drawLune = winTerre->drawLune = val_drawLune;
    winStars->drawPlanetes = val_drawPlanetes;
    winStars->drawSatellites = winTerre->drawSatellites = val_drawSatellites;
    precision = val_precision;
    winTerre->gainDist = val_gainDist;
    winTerre->gainAngle = val_gainAngle;
    winTerre->aziScene = val_aziScene;
    winTerre->sitScene = val_sitScene;
    winTerre->perspective = val_perspective;

    if (creneau > periode) creneau = periode;

    if (accel < 0) {dtsim = -creneau/accel; accel = dtsim/(dtime_ms*0.001);}
    else dtsim = accel*dtime_ms*0.001;
    if (dtsim > creneau) {dtsim = creneau; accel = dtsim/(dtime_ms*0.001);}

    synchroTimeDialogSlot(); // Met en concordance les paramètres temporels (pas, accélération, etc)
    timeDialog->init(); // Met à jour le temps et les intervalles du timeDialog
    winStars->confStarsDialog->init(); // met à jour le confStarsDialog
    winStars->majVisee(); // Met à jour les paramètres d'observation du ciel étoilé
    winStars->majColors(); // Met à jour la couleur des étoiles et du ciel (noir/blanc, ou autre..)
    if (winTerre->modePtVue != (ModePtVue) val_terrePDV)
    {
        winTerre->modePtVue = (ModePtVue) val_terrePDV;
        winTerre->show();
        winTerre->setProjection();
    }
    update(tdeb); // Met à jour les positions et les vues.
}

