#include "orbitgeo.h"
#include "math.h"
#include <stdio.h>
#include "terre.h"
#include "../mathutil/elem.h"

/** Construction d'une orbite de satellite à partir de ses elements orbitaux
	 *  en unites S.I.
   @param n0		mouvement moyen(radians/s)
   @param e			excentricite
   @param i0		inclinaison de l'orbite (radians)
   @param om0		argument du perigee (radians)
   @param gdOm0		ascension droite du noeud ascendant (radians)
	 */	
OrbitGeo::OrbitGeo(double n0, double e, double i0, double om0, double Om0)
{
    set(n0, e, i0, om0, Om0);
}

/**
 * Calcule l'ensemble coherent de tous les parametres orbitaux
 * memorises, a partir de "n0, e, i0, om0 et Om0".
 */
void OrbitGeo::set(double n0_, double e, double i0_, double om0_, double Om0)
{
    n0 = n0_;
    e0 = e;
    i0 = i0_;
    om0 = om0_;
    gdOm0 = Om0;
    double a1 = pow(Terre::GM/(n0*n0),1./3.);
    double temp, rap, r, rmin;
    int qrep;
    double ume2 = 1.- e*e;
    ci0 = cos(i0);
    si0 = sin(i0);

    m3_HN[0] = 1.;  m3_HN[3] =  0.; m3_HN[6] =   0.;
    m3_HN[1] = 0.;  m3_HN[4] = ci0; m3_HN[7] = -si0;
    m3_HN[2] = 0.;  m3_HN[5] = si0; m3_HN[8] =  ci0;

    a0 = a1;
    double p = a0*ume2;
    double rT2J2sp2 = Terre::r2J2/(p*p);
    delta1 =  0.75*rT2J2sp2*sqrt(ume2)*(3.*ci0*ci0-1.); // Pour SGP
    temp = rT2J2sp2*n0;
    gdOmP = -1.5*temp*ci0;
    omP = 0.75*temp*(5.*ci0*ci0-1.);
    nnod = n0 + omP;
    Tn0 = 2*M_PI/n0;
    Tnod = 2*M_PI/nnod;
    // Calcul de la periode de repetitivite
    rap = 86400./Tnod;
    qrep = 1;
    rmin = fabs(qrep*rap - floor(qrep*rap + 0.5));
    for (int i = 2; i < 32; i++)
    {
        r = fabs(i*rap - floor(i*rap + 0.5));
        if (r < rmin)
        {
            rmin = r;
            qrep = i;
        }
    }
    Qrep = qrep;
    Krep = (int) floor(Qrep*rap + 0.5);
}

/**
  * Impression des parametres orbitaux
*/
void OrbitGeo::print()
{
    double temp;
    printf("Inclinaison        : %.4lf dg\n", toDegrees(i0));
    printf("Excentricite       : %lg\n", e0);
    printf("Asc. Droite NA.    : %.4lf dg\n", toDegrees(gdOm0));
    printf("Arg. perigee       : %.4lf dg\n", toDegrees(om0));
    printf("Mouvement moyen    : %lf tr/j\n", 86400.*n0*0.5/M_PI);
    printf("Demi-grand axe     : %.6lf km\n", a0/1000.);
    printf("Altitude           : %.6lf km\n", (a0-Terre::rT)/1000.);
    printf("Vit. AscDrNA       : %lg rd/s\n", gdOmP);
    printf("Vit. Arg.Perigee   : %lg rd/s\n", omP);
    temp = 2*M_PI/Terre::antrop2s;
    double pcsync = 100.*fabs((gdOmP - temp)/temp); // Pourcentage d'erreur
    printf("Orbite heliosyncrone a %.2lf %% pres\n", pcsync);
    printf("La periode de repetitivite est de %d  jours\n", Qrep);
    printf("Elle met en jeu %d orbites\n", Krep);
    printf("Le rendez-vous a lieu a %.1lf secondes pres\n", Krep*Tnod - Qrep*86400.);
}

