#include "reperes.h"
#include "../mathutil/rotation.h"
#include "../mathutil/sp3.h"
#include <math.h>

static const double  torad = (M_PI/180.);

// Pour les repères écliptiques
double Reperes::lastsJe1 = 0.;
double Reperes::lastsJe2 = 0.;
double Reperes::eta = 0.;
double Reperes::seta = 0.;
double Reperes::ceta = 1.;
double Reperes::sigma = 3.052168685;
double Reperes::esse = 0.;
double Reperes::m3_EE21[9] = {1.,0.,0.,0.,1.,0.,0.,0.,1.};
double Reperes::deltaEpsilon = 0.;
double Reperes::deltaPsi = 0.;

// Pour les repères equatoriaux
double Reperes::lastsJc1 = 0.;
double Reperes::lastsJc2 = 0.;
double Reperes::zede = 0.;
double Reperes::teta = 0.;
double Reperes::steta = 0.;
double Reperes::cteta = 1.;
double Reperes::zeta = 0.;
double Reperes::m3_CC21[9] = {1.,0.,0.,0.,1.,0.,0.,0.,1.};

// Liaison entre les deux
const double Reperes::varPrecSiecle = 0.0244; //Erreur de precession en radian par siècle
const double Reperes::errBradleyMax = 9.7e-5; // Erreur de bradley max (20" en 18.61 ans) conv. en radians
const double Reperes::errBradSiecle = 5.22e-4; // Erreur de Bradley par siecle
const double Reperes::varObliqSiecle = 2.269e-4; // Variation Obliquité par siècle

double Reperes::lastsJObliq = -1.e+20; // dernier calc obliquité
double Reperes::lastCorNut = -1.e+20; // dernière correction de Bradley
/** Obliquite de l'ecliptique en degres */
const double Reperes::eps0_dg = 23.439291111;
double Reperes::epsilon_dg = Reperes::eps0_dg;
double Reperes::epsilon = torad*Reperes::epsilon_dg;
double Reperes::coseps = cos(Reperes::epsilon);
double Reperes::sineps = sin(Reperes::epsilon);
/** Matrice passage Re -> Rc */
double Reperes::m3_CE[9] = {1.,0.,0., 0., Reperes::coseps, Reperes::sineps,
                          0. , -Reperes::sineps, Reperes::coseps};
/** Matrice passage Rcv -> Rcm */
double Reperes::m3_CCvm[9] = {1.,0.,0.,0.,1.,0.,0.,0.,1.};

/** Mise à jour de l'obliquité
      @param sj 	Instant du calcul en siecles juliens / J2000 (TE ou TU)
 *
 *  On met à jour :
 *      epsilon_dg : obliquité moyenne à la date t_j en degré
 *      epsilon    : idem en radian
 *  coseps, sineps : lignes trigo.
 *      m3_CE      : matrice de passage Celeste -> Ecliptique à la date t_j
 *      lastsJObliq: dernier instant de calcul en siècle/J2000
 */
void Reperes::obliquite(double T)
{
    epsilon_dg = eps0_dg + T*(-0.013004166667 +T*(-0.000000163889 + T*0.000000503611));
    epsilon = torad*epsilon_dg;
    // mise a jour matrice passage écliptique moyen de la date -> céleste moyen de la date en cas de besoin externe
    m3_CE[4] = m3_CE[8] = coseps = cos(epsilon);
    m3_CE[5] = sineps = sin(epsilon);
    m3_CE[7] = -sineps;
    lastsJObliq = T;
}


// Mise a jour du changement de coordonnnees ecliptiques moyennes
// Entrees :
//   sJ1 : premiere date en siècles juliens depuis J2000
//         c'est a priori celle a laquelle on possede les coordonnees de l'astre
//   sJ2 : deuxième date en siècles juliens depuis J2000
//         c'est a priori celle a laquelle on desire les coordonnees de l'astre
// Mise à jour des angles eta, sigma et esse et de la matrice m3_EE21
void Reperes::majEclipMoy(double sJ1,double sJ2)
{
    double dSJ = (sJ2-sJ1);   // delta t entre dates initiale et finale en siecles juliens
    double a = 47.0029 + sJ1*(-0.06603 + 0.000598*sJ1);
    double b = -0.03302+0.000598*sJ1;
    eta = dSJ*(a+dSJ*(b+dSJ*0.00006)); // en secondes angulaires
    eta = (eta/3600)*M_PI/180.;    // en radians
    a = 174.876384*3600 + sJ1*(3289.4789+0.60622*sJ1);
    b = 869.8089+0.50491*sJ1;
    sigma = a+dSJ*(-b+dSJ*0.03536);    // en secondes angulaires
    sigma /=3600; // en degres
    sigma *= M_PI/180.;     // en radians
    a = 5029.0966 + sJ1*(2.22226 - 0.000042*sJ1);
    b = 1.11113-0.000042*sJ1;
    esse = dSJ*(a+dSJ*(b-dSJ*0.000006)); // en secondes angulaires
    esse = (esse/3600)*M_PI/180.;    // en radians
    ceta = cos(eta); seta = sin(eta);
    // Mise a jour de la matrice de passage m3_EE21, telle que (V)_2 = m3_EE21 (V)_1
    double ma[9],mb[9],mc[9];
    m3m3(m3_EE21, m3m3(ma, mrotz(ma, esse+sigma), mrotx(mb,-eta)), mrotz(mc,-sigma));
    lastsJe1 = sJ1; lastsJe2 = sJ2;
}

// Mise a jour du changement de coordonnnees équatoriales moyennes
// Entrees :
//   sJ1 : premiere date en siècles juliens depuis J2000
//         c'est a priori celle a laquelle on possede les coordonnees de l'astre
//   sJ2 : deuxième date en siècles juliens depuis J2000
//         c'est a priori celle a laquelle on desire les coordonnees de l'astre
// Mise à jour des angles zeta, zede et teta et de la matrice m3_CC21
void Reperes::majEquatMoy(double sJ1,double sJ2)
{
    double dSJ = (sJ2-sJ1);   // delta t entre dates initiale et finale en siecles juliens
    double a = 2306.2181 + sJ1*(1.39656 - 0.000139*sJ1);
    double b = 0.30188-0.000344*sJ1;
    zeta = dSJ*(a+dSJ*(b+dSJ*0.017998)); // en secondes angulaires
    zeta = (zeta/3600)*M_PI/180.;    // en radians
    b = 1.09468+0.000066*sJ1;
    zede = dSJ*(a+dSJ*(b+dSJ*0.018203));     // en secondes angulaires
    zede = (zede/3600)*M_PI/180.;    // en radians
    a = 2004.3109 - sJ1*(0.85330 + 0.000217*sJ1);
    b = 0.42665+0.000217*sJ1;
    teta = dSJ*(a+dSJ*(-b-dSJ*0.041883)); // en secondes angulaires
    teta = (teta/3600)*M_PI/180.;    // en radians
    cteta = cos(teta); steta = sin(teta);
    // Mise a jour de la matrice de passage m3_CC21, telle que (V)_2 = m3_CC21 (V)_1
    double ma[9],mb[9],mc[9];
    m3m3(m3_CC21, m3m3(ma, mrotz(ma, zede), mroty(mb,-teta)), mrotz(mc,zeta));
    lastsJc1 = sJ1; lastsJc2 = sJ2;
}

// Conversion de coordonnnees ecliptiques moyennes d'une etoile d'une date vers une autre date
//   lConnu, bConnu : coord. ecli. connues en RADIANS
//   lInc, bInc, sinbInc : coord. ecli. cherchées en RADIAN et SINUS
//   direct :
//      si vrai : (lConnu, bConnu) sont les coordonnées initiales et (lInc, bInc) les coord. finales cherchées
//      si faux : (lConnu, bConnu) sont les coordonnées finales et (lInc, bInc) les coord. initiales cherchées
void Reperes::chgCoordsEclip(double *lInc, double *bInc, double *sinbInc, double lConnu, double bConnu, bool direct)
{
    double l0s, cl0s, sl0s, cb0, sb0, A, B, C;
    if (direct)
    {
        l0s = lConnu - sigma;
        cl0s = cos(l0s);
        sl0s = sin(l0s);
        cb0 = cos(bConnu);
        sb0 = sin(bConnu);
        A = cb0*cl0s;
        B = ceta*cb0*sl0s + seta*sb0;
        C = -seta*cb0*sl0s + ceta*sb0;
        *lInc = esse + sigma + atan2(B,A);
        *bInc = asin (C);
        *sinbInc = C;
    }
    else
    {
        l0s = lConnu - sigma - esse;
        cl0s = cos(l0s);
        sl0s = sin(l0s);
        cb0 = cos(bConnu);
        sb0 = sin(bConnu);
        A = cb0*cl0s;
        B = ceta*cb0*sl0s - seta*sb0;
        C = seta*cb0*sl0s + ceta*sb0;
        *lInc = sigma + atan2(B,A);
        *bInc = asin (C);
        *sinbInc = C;
    }
    return;
}

// Conversion de coordonnnees équatoriales moyennes d'une etoile d'une date vers une autre date
//   aConnu, dConnu : coord. equa. connues en RADIANS
//   aInc, dInc, sindInc : coord. equa. cherchées en RADIAN et sinus
//   direct :
//      si vrai : (aConnu, dConnu) sont les coordonnées initiales et (aInc, dInc) les coord. finales cherchées
//      si faux : (aConnu, dConnu) sont les coordonnées finales et (aInc, dInc) les coord. initiales cherchées
void Reperes::chgCoordsEquat(double *aInc, double *dInc, double *sindInc, double aConnu, double dConnu, bool direct)
{
    double A,B,C,ca0z, sa0z, cd0, sd0;
    if (direct)
    {
        aConnu += zeta;
        ca0z = cos(aConnu); sa0z = sin(aConnu);
        cd0 = cos(dConnu); sd0 = sin(dConnu);
        A = cteta*cd0*ca0z - steta*sd0;
        B = cd0*sa0z;
        C = steta*cd0*ca0z + cteta*sd0;
        *aInc = zede + atan2(B,A);
        *dInc = asin (C);
        *sindInc = C;
    }
    else
    {
        aConnu -= zede;
        ca0z = cos(aConnu); sa0z = sin(aConnu);
        cd0 = cos(dConnu); sd0 = sin(dConnu);
        A = cteta*cd0*ca0z + steta*sd0;
        B = cd0*sa0z;
        C = -steta*cd0*ca0z + cteta*sd0;
        *aInc = -zeta + atan2(B,A);
        *dInc = asin (C);
        *sindInc = C;
    }
    return;
}

/** Mise à jour pour correction nutation (mouv. de Bradley)
 * Entrees :
 *    sJ 	Instant du calcul en siècles juliens / J2000 (TE ou TU)
 *  precismax : vrai si précision max demandée
 * Mises à jour de :
 *    deltaEpsilon : variation d'obliquité de l'axe céleste
 *    deltaPsi     : variation de l'axe céleste en longitude écliptique
 *    m3_CCvm      : matrice de passage repère vrai -> repère moyen
 *    tsNutation   : memo dernier instant de calcul, en siècle/J2000
 */
void Reperes::majEclipVrai(double sJ, bool precisionmax)
{
    double T = sJ;
    double L  = (280.4665 + T*( 36000.7698 + T*0.000303))*torad;
    double Lp = (218.3164 + T*(481267.8812 - T*0.001599))*torad;
    double M  = (357.5291 + T*( 35999.0503 - T*0.000154))*torad;
    double O  = (125.0443 + T*( -1934.1363 + T*0.002075))*torad;
    double dL = 2.*L, dLp = 2.*Lp, dO = 2.*O;
    deltaPsi = -(17.1996+0.01742*T)*sin(O)
               -( 1.3187+0.00016*T)*sin(dL)
               - 0.2274*sin(dLp)
               + 0.2062*sin(dO)
               + (0.1426-0.00034*T)*sin(M);
    deltaEpsilon = (9.2025+0.00089*T)*cos(O)
               + (0.5736-0.00031*T)*cos(dL)
               + 0.0977*cos(dLp)
               - 0.0895*cos(dO);
    if (precisionmax)
    {
        double Mp = (134.9634 + T*(477198.8675 + T*0.008721))*torad;
        deltaPsi +=    0.0712*sin(Mp)
                    - (0.0517 - 0.00012*T)*sin(dL+M)
                    - 0.0386*sin(dLp-O)
                    - 0.0301*sin(dLp+Mp)
                    + 0.0217*sin(dL-M)
                    - 0.0158*sin(dL-dLp+Mp)
                    + 0.0129*sin(dL-O)
                    + 0.0123*sin(dLp-Mp);

        deltaEpsilon +=  0.0224*cos(dL+M)
                       + 0.0200*cos(dLp-O)
                       + 0.0129*cos(dLp+Mp)
                       - 0.0095*cos(dL-M)
                       - 0.0070*cos(dL-O);
    }
    deltaPsi *= torad/3600.;
    deltaEpsilon *=  torad/3600.;

    m3_CCvm[1] = deltaPsi*coseps;
    m3_CCvm[2] = deltaPsi*sineps;
    m3_CCvm[3] = -m3_CCvm[1];
    m3_CCvm[6] = -m3_CCvm[2];
    m3_CCvm[5] = deltaEpsilon;
    m3_CCvm[7] = -deltaEpsilon;
    lastCorNut = T;
    return;
}

/** Calcul de la correction des coordonnées équatoriales due à la nutation (mouv. de Bradley)
 *  Suppose l'appel au préalable de obliquite et majEclipVrai pour la mise-à-jour de
 *  l'obliquité Epsilon et des variations deltaPsi et deltaEpsilon
 * Entrees :
 *    alf, dec : Coordonnées équatoriales à corriger
 * Sorties :
 *    deltaAlf, deltaDec : corrections à appliquer
 * Remarque : Une autre approche consiste à appliquer la matrice m3_CCvm aux composantes
 *    de la direction de l'objet céleste.
 */
void Reperes::deltaCoordEquatNutation(double *deltaAlf, double *deltaDec, double alf, double dec)
{
   double sina = sin(alf), cosa = cos(alf);
   double tand = tan(dec);

    *deltaAlf = (coseps + sineps*sina*tand)*deltaPsi - cosa*tand*deltaEpsilon;
    *deltaDec = sineps*cosa*deltaPsi + sina*deltaEpsilon;
    return;
}
