#include "terre.h"
#include "reperes.h"
#include "../mathutil/sp3.h"
#include "../mathutil/elem.h"
#include "../utils/cjulien/julien.h"
#include <math.h>

static const double  torad = M_PI/180.;

// ATTENTION : Les matrices 3x3 rangées en [9], le sont par colonne par colonne.

// CLASSE STATIQUE. Pas besoinde créer d'instance.

/* ------------------------
   Constantes de la classe
 * ------------------------ */
const double Terre::siecleJul_j = 36525.;
const double Terre::siecleJul_s = Terre::siecleJul_j*86400.;

const double Terre::raySol = 696342.0e03;

/** Unite astronomique (valeur IERS 2003) en m */
const double Terre::UA = 149597870691.;

/** excentricite nominale de l'orbite Terre */
const double Terre::e0 = 0.016708634;

/** excentricite prise encompte de l'orbite Terre non nulle si flagcpx >= 1 */
double Terre::e =  Terre::e0;

/** Annee siderale en jours */
const double Terre::anside2j = 365.25636305 ;

/** Heure siderale Greenwich a J2000.0 en degres */
const double Terre::hsg0_dg = 280.46061837;

/** Heure siderale Greenwich a J2000.0 en radians */
const double Terre::hsg0 = Terre::hsg0_dg * M_PI/180.;

/** Heure siderale Greenwich a J2000.0 en secondes (67310.54840) */
const double Terre::hsg0_s = Terre::hsg0_dg*86400./360.;

/** Annee tropique en jours (differente Annee siderale si flagcpx >= 1) */
double Terre::antrop2j = 365.2421904 ;

/** Annee tropique en secondes (differente Annee siderale si flagcpx >= 1) */
double Terre::antrop2s = Terre::antrop2j * 86400.;

/** Vitesse de rotation de la Terre en rd/jour */
const double Terre::omegaT = 2.*M_PI*(Terre::antrop2j+1.)/Terre::antrop2j;

/** Rayon equatorial Terre en m (valeur IERS 2003) */
const double Terre::rT = 6378136.6;

/** Cte. gravitation de la Terre en S.I. (valeur IERS 2003) */
const double Terre::GM = 3.986004418e14;

/** Harmonique spherique J2 du potentiel Terrestre (valeur IERS 2003) */
const double Terre::J2 = 0.0010826359;

/** Harmonique spherique J3 du potentiel Terrestre */
const double Terre::J3 = -0.00000254;

/**    */
const double Terre::r2J2 = Terre::rT*Terre::rT*Terre::J2;

/** Aplatissement de l'ellipsoide terrestre (valeur IERS 2003) pris nul si flagcpx < 2 */
double Terre::f = 1/98.25642;

/** Parametre p de l'ellipsoide terrestre */
double Terre::pT = Terre::rT*(1.-Terre::f)*(1.-Terre::f);

/** Carre de l'excentricite de l'ellipsoide terrestre */
double Terre::eT2 = Terre::f*(2.-Terre::f);

/** Heure siderale Greenwich a la date t en radians */
double Terre::hsg = 0.;
double Terre::coshsg = 1.;
double Terre::sinhsg = 0.;

/** Matrice passage Rc -> Rg */
double Terre::m3_CG[9] = {1.,0.,0.,0.,1.,0.,0.,0.,1.};

/** Distance du Soleil a la Terre */
double Terre::distTO = Terre::UA;

/** Cosinus directeurs vecteur Terre-Soleil dans Rc */
double Terre::v3_uTO_c[3] = {1., 0., 0.};

/** Ascension droite du Soleil dans Rc */
double Terre::solAscen = 0.;

/** Declinaison du Soleil dans Rc */
double Terre::solDecli = 0.;

/** Longitude écliptique du Soleil dans Reclip */
double Terre::solLongiEclip = 0.;
double Terre::cosLE = 1.;
double Terre::sinLE = 0.;

double Terre::solLongiPeri = 0.; // valeur sans sens !

/** Latitude du Soleil dans Rg */
double Terre::solLati = 0.;

/** Observateur sur Terre */
double Terre::longobs = 0.;
double Terre::latobs = 0.;
double Terre::m3_GA[9] = {1., 0., 0., 0., 1., 0., 0., 0., 1.};
double Terre::m3_CA[9] = {1., 0., 0., 0., 1., 0., 0., 0., 1.};
double Terre::v3_GA[3] = {Terre::rT, 0., 0.};


// Instants improbables
double Terre::tjOrien = -1.e+20;
double Terre::tjSol = -1.e+20;

/** Position du Soleil
      @param te_j 	Instant du calcul TE en jours / J2000
      @param prec 	précision désirée pour les résultats
*/
void Terre::posSoleil(double te_j, double precision)
{
#define errJourRev 0.0172 // variation angulaire par jour
    // Position du Soleil
    if (fabs(tjSol - te_j)*errJourRev >= precision)
    {
        tjSol = te_j;
        double L, C, M;
        // correction retard vitesse lumiere
        double t_j = te_j - distTO/(3.E8*86400.);
        //Date T en siecles juliens depuis J2000.0
        double T = t_j/siecleJul_j;
        // Anomalie Moyenne
        M = 357.52911 + T*(35999.05029 - T*(0.0001537 + T*0.0000033));
        M = toRadians(realmod(M,360.));
        // Longitude ecliptique moyenne
        L = 280.46646 + T*(36000.76983 + T*0.0003032);
        L = toRadians(realmod(L,360.));
        solLongiPeri = L - M; // Pour calcul precis aberration annuelle
        C = toRadians((1.914602-T*(0.004817+T*0.000014))*sin(M)
                      + (0.019993-0.000101*T)*sin(2*M)
                      + 0.000289*sin(3*M));
        // Anomalie vraie (M+C)
        // Rayon vecteur
        e = e0 - T*(0.000042037+T*0.0000001267);
        distTO = UA*1.000001018*(1. - e*e)/(1. + e*cos(M + C));
        // Longitude géométrique écliptique du Soleil relative à l'equinoxe moyen de la date
        solLongiEclip = L + C;
        sinLE = sin(solLongiEclip);
        cosLE = cos(solLongiEclip);
        // Calcul obliquité
        if (fabs(Reperes::lastsJObliq - T)*Reperes::varObliqSiecle >= precision) Reperes::obliquite(T);
        // La latitude ecliptique relative à l'equinoxe moyen de la date tjs < à 1.2 seconde d'arcs est négligée
        double u = cosLE;
        double v = sinLE*Reperes::coseps;
        double w = sinLE*Reperes::sineps;
        v3_uTO_c[0] = u; v3_uTO_c[1] = v; v3_uTO_c[2] = w;
        solAscen = atan2(v, u);
        solDecli = solLati = asin(w);
    }
    return;
}


/** Orientation de la Terre dans le repère moyens de la date
      @param tu_j 	Instant du calcul TU en jours / J2000
      @param prec 	précision désirée pour les résultats
*/
void Terre::orientation(double tu_j, double precision)
{
#define errJourRot 6.28318 // variation angulaire par jour

    if (fabs(tjOrien - tu_j)*errJourRot >= precision)
    {
        double T = tu_j/36525.; // siècles juliens
        // Correction éventuelle de la Nutation -> repère céleste vrai)
        if(precision < Reperes::errBradleyMax && fabs(Reperes::lastCorNut - T)*Reperes::errBradSiecle >= precision)
        {
            // Maj éventuelle prélable de l'obliquité
            if(fabs(Reperes::lastsJObliq - T)*Reperes::varObliqSiecle >= precision)
                Reperes::obliquite(T);
            Reperes::majEclipVrai(T, precision < 1.e-6);
        }
        // heure siderale de Greenwich (SANS LA CORRECTION DE TEMPS APPARENT = Reperes::m3_CCvm[1])
        hsg = hsg0 + omegaT*tu_j + (6.770708127e-6 - 4.50872966e-010*T)*T*T; // + Reperes::m3_CCvm[1];
        coshsg = cos(hsg);
        sinhsg =  sin(hsg);
        m3_CG[0] = m3_CG[4] = coshsg;
        m3_CG[1] = sinhsg;
        m3_CG[3] = -sinhsg;
        m3_CG[8] = 1.;
        m3_CG[6] = m3_CG[7] = m3_CG[2] = m3_CG[5] = 0.;
        m3m3(m3_CA, m3_CG, m3_GA); // En l'abscence de correction
        tm3m3(m3_CA, Reperes::m3_CCvm, m3_CA); // Avec correction de la nutation
        tjOrien = tu_j;
    }
    return;
}

// Mise a jour longitude latitude
void Terre::setLongiLati(double longi, double lati)
{
    // Matrice d'attitude du repère local
    longobs = longi;
    latobs = lati;
    lolaAstroTom3(m3_GA, longi, lati);
    cgeod2ccart(v3_GA,longi, lati,0.);
}

/** Lever et coucher du soleil, heure et azimut
     @param tu_j 	Instant du calcul en jours / J2000 (TU)
     @param te_j 	Instant du calcul en jours / J2000 (TT)
     @param longiA  Longitude du lieu d'observation
     @param latiA   Latitude du lieu d'observation
     @param lnc 0 : sans raffinement, 1 avec raffinement pour lever, 2 avec raffinement pour coucher
     @param pt      pointeur struct resultat
*/
void Terre::sunhzero(tsunhzero *pt, double tu_j, double te_j, double longiA, double latiA, int lnc)
{

    for(;;)
    {
        // jours est entier a midi
        posSoleil(te_j, 1.e-5);
        // Calcul écart angulaire alphaA - alphaSol
        double cosEcart = -tan(latiA)*tan(solDecli);
        if (cosEcart > 1.) cosEcart = 1.; if (cosEcart < -1.) cosEcart = -1.;
        double ecartAplha = acos(cosEcart);
        // Calcul Equation du temps (Ecart angulaire soleil vrai - soleil moyen)
        double T = tu_j/36525;
        double tplat = floor(tu_j + 0.5*longiA/M_PI + 0.5); // date courante, en j/J2000 ramenée à 12h locale maintenue de 0 à 23h59
        double alphaJA = hsg0 + omegaT*tplat + (6.77071e-6 - 4.50873e-010*T)*T*T;
        alphaJA = realmod2(alphaJA, 2.*M_PI);
        pt->leverLoc = (solAscen - ecartAplha - alphaJA)/omegaT; // écart temporel par rapport a midi
        pt->coucherLoc =  (solAscen + ecartAplha - alphaJA)/omegaT; // écart temporel par rapport a midi
        pt->leverUT = tplat + pt->leverLoc - longiA/(2.*M_PI);
        pt->coucherUT = tplat + pt->coucherLoc - longiA/(2.*M_PI);

        double cosAzi = -sin(solDecli)/cos(longiA);
        if (cosAzi > 1.) cosAzi = 1.; if (cosAzi < -1.) cosAzi = -1.;
        double azi = fabs(acos(cosAzi));
        pt->azilever = azi;
        pt->azicoucher = -azi;
        if (lnc == 0) break;
        if (lnc == 1) tu_j = pt->leverUT; // On raffine une fois à l'instant du lever
        else tu_j = pt->coucherUT; // On raffine une fois à l'instant du coucher
        lnc = 0; // On sort au coup suivant
    }
}


/** Conversion coordonnees geodesiques -> cartesiennes
      @param V 			Sortie V[3] : x,y,z
      @param longi 		longitude
      @param lati 		latitude geodesique
      @param alti 		altitude geodesique
         */
void Terre::cgeod2ccart(double *V, double longi, double lati, double alti)
{
    double sf = sin(lati);
    double cf = cos(lati);
    double unr = sqrt(1. - eT2*sf*sf);
    double v = (rT/unr + alti)*cf;
    V[0] = v*cos(longi);
    V[1] = v*sin(longi);
    V[2] = (pT/unr + alti)*sf;
    return;
}


/** Conversion coordonnees cartesiennes -> geodesiques
      @param V 	  Sortie V[0:2] : longi, lati, h
      @param M	  double [3] des coordonnees cartesiennes x,y,z
      @param eps   precision relative desiree
         */
void Terre::ccart2cgeod(double *V, double *M, double eps)
{
    double lambda = atan2(M[1], M[0]);
    double w = M[0]*cos(lambda) + M[1]*sin(lambda);
    double h = 0., phi = 0., sphi = 0., temp = 1., rr, rp;
    double oldphi, oldh, dphi;
    double eps2 = rT * eps;
    for (int i = 0; i < 10; i++) // Converge habituellement en 2 iterations
    {
        rr = rT/temp;
        rp = pT/temp;
        oldphi = phi;
        phi = atan2(M[2]/(rp+h), w/(rr+h));
        sphi = sin(phi);
        temp = sqrt(1.-eT2*sphi*sphi);
        oldh = h;
        h = w*cos(phi) + M[2]*sphi -rT*temp;
        dphi = realmod2(phi-oldphi,2.*M_PI);
        if (fabs(dphi) < eps && fabs(h-oldh) < eps2) break;
    }
    V[0] = lambda;
    V[1] = phi;
    V[2] = h;
}

/** Impression position Soleil dans Rc .  */
void Terre::print()
{
    double Vbid[3];
    scav3(Vbid, distTO*1.e-3, v3_uTO_c);
    fprintv3("Position Soleil dans Rc (en km) : ", Vbid, stdout, "%11.0lf");
}

#if 1
// POUR FAIRE UN TEST
void testTerre()
{
    double t = 0.;
    double T, M_Sol, L_Sol, C_Sol, sin_Lsol, cos_Lsol ;
    double V_Sol[3];

    const double deltaTE = 63.61;
    const double siecle_s = Terre::siecleJul_s;
    const double eps_Sol = toRadians(23.43928);
    const double M0_Sol = toRadians(357.52543);
    const double Mprime_Sol = toRadians(35999.04944);
    const double L0_Sol = toRadians(280.4659);
    const double Lprime_Sol = toRadians(36000.769525);
    const double C1_Sol = 1.914657;
    const double C2_Sol = 0.019994;
    const double C3_Sol = 0.000293;

    Terre::posSoleil(t/86400., 1.e-6);
    Terre::print();
    fprintv3("Direction Soleil dans Rc : ", Terre::v3_uTO_c, stdout, "%7.3lf");

    fprintf(stdout, "Verification :");
    t = 1000.;
    T = (t + deltaTE)/siecle_s; // Siecles juliens depuis J2000.0
    M_Sol = M0_Sol + Mprime_Sol*T; // Anomalie Moyenne
    L_Sol = L0_Sol + Lprime_Sol*T; // Longitude ecliptique moyenne
    C_Sol = toRadians(C1_Sol*sin(M_Sol)
                      + C2_Sol*sin(2*M_Sol)
                      + C3_Sol*sin(3*M_Sol));
    L_Sol = L_Sol + C_Sol;
    sin_Lsol = sin(L_Sol);
    cos_Lsol = cos(L_Sol);
    V_Sol[0] = cos_Lsol;
    V_Sol[1] = sin_Lsol*cos(eps_Sol);
    V_Sol[2] = sin_Lsol*sin(eps_Sol);
    fprintv3("Direction Soleil dans Rc : ", V_Sol, stdout, "%7.3lf");
    return;
}
#endif
