#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <float.h>
#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif
#include "rotation.h"
#include "sp3.h"

#define MY_EPSREL 1.E-14

static int printformat=1113;
static char strfmt[10]="%11.3lg";
/**********************************************************************/
/*        UTILITAIRES ROTATION ET DEPLACEMENTS DANS R3                */
/**********************************************************************/
/*---------------------------------------------------------------------
	Génération matrice de rotation d'angles d'euler
		Précession (Oz) , Nutation (Ox1) Rotation propre (Oz2)
	----------------------------------------------------------

	Entrees:
		double *a	: Adresse matrice affectée
		double *euler	: Adresse des angles Psi,Tet,Phi (radians)

	Sorties:
		a est calculée
		Retour		: Adresse de la matrice affectée
----------------------------------------------------------------------*/
double *euler2mrot(double *a, double *euler)
{
double cp, sp, ct, st, cf, sf ;
double ctsf, ctcf ;

    cp = cos (euler[0]) ; ct = cos (euler[1]) ; cf = cos (euler[2]) ;
    sp = sin (euler[0]) ; st = sin (euler[1]) ; sf = sin (euler[2]) ;
    ctsf = ct * sf ; ctcf = ct * cf ;

    a[0] = cp*cf - sp*ctsf ; a[3] = -cp*sf - sp*ctcf ; a[6] =  sp*st ;
    a[1] = sp*cf + cp*ctsf ; a[4] = -sp*sf + cp*ctcf ; a[7] = -cp*st ;
    a[2] =           st*sf ; a[5] =            st*cf ; a[8] =     ct ;
	return a;
}

/*---------------------------------------------------------------------
	Génération matrice de rotation d'angles aéronautique
				   Cap (Oz) , Assiette (Oy1) Roulis (Ox2)
	-----------------------------------------------------

	Entrees:
		double *a	: Adresse matrice affectée
		double *carol	: Adresse des angles C,A,R (radians)

	Sorties:
		a est calculée
		Retour		: Adresse de la matrice affectée
----------------------------------------------------------------------*/
double *carol2mrot(double *a, double *carol)
{
double cc, sc, ca, sa, cr, sr ;
double sasr, sacr ;

    cc = cos (carol[0]) ; ca = cos (carol[1]) ; cr = cos (carol[2]) ;
    sc = sin (carol[0]) ; sa = sin (carol[1]) ; sr = sin (carol[2]) ;
    sasr = sa * sr ; sacr = sa * cr ;

    a[0] = cc*ca  ; a[3] = cc*sasr - sc*cr ; a[6] = cc*sacr + sc*sr ;
    a[1] = sc*ca  ; a[4] = sc*sasr + cc*cr ; a[7] = sc*sacr - cc*sr ;
    a[2] =   -sa  ; a[5] =    ca*sr        ; a[8] =    ca*cr        ;
	return a;
}

/*---------------------------------------------------------------------
	Génération matrice de rotation d'angles
								 L (Oz) , T (Ox1) R (Oy2)
	-----------------------------------------------------

	Entrees:
		double *a	: Adresse matrice affectée
		double *ltr	: Adresse des angles ltr (radians)

	Sorties:
		a est calculée
		Retour		: Adresse de la matrice affectée
----------------------------------------------------------------------*/
double *ltr2mrot(double *a, double *ltr)
{
double cl, sl, ct, st, cr, sr ;
double slst, clst ;

    cl = cos (ltr[0]) ; ct = cos (ltr[1]) ; cr = cos (ltr[2]) ;
    sl = sin (ltr[0]) ; st = sin (ltr[1]) ; sr = sin (ltr[2]) ;
    slst = sl * st ; clst = cl * st ;

    a[0] = cl*cr - slst*sr ; a[3] = -sl*ct ; a[6] = cl*sr + slst*cr ;
    a[1] = sl*cr + clst*sr ; a[4] =  cl*ct ; a[7] = sl*sr - clst*cr ;
    a[2] =          -ct*sr ; a[5] =     st ; a[8] =           ct*cr ;
	return a;
}

/*----------------------------------------------------------------------
    GENERATION MATRICE ROTATION :  na2mrot(a,k,g)
    ---------------------------------------------
    Entrees:
       int k        : Numero axe de rotation (1=X, 2=Y ou 3=Z)
       double g     : Angle rotation en radians
	   double *a	: Adresse matrice affectée

    Sorties:
    Sorties:
		a est calculée
		Retour		: Adresse de la matrice affectée

----------------------------------------------------------------------*/
double *na2mrot(double* a, int k, double g)
{
double s, c;
    c = cos(g);
    s = sin(g);
    switch (k)
    {
	case 1:
	    a[0] = 1.;
	    a[1] = a[2] = a[3] = a[6] = 0.;
	    a[4] = a[8]= c;
	    a[5] = s;
	    a[7] = -s;
	    break ;
	case 2:
	    a[4] = 1.;
	    a[1] = a[3] = a[5] = a[7] = 0.;
	    a[0] = a[8]= c;
	    a[6] = s;
	    a[2] = -s;
	    break ;
	case 3:
            a[8] = 1.;
            a[2] = a[5] = a[6] = a[7] = 0.;
            a[0] = a[4]= c;
            a[1] = s;
	    a[3] = -s;
	    break ;
    }
    return a;
}

/*----------------------------------------------------------------------
    GENERATION MATRICE ROTATION :  angs2mrot(m,angs,i,j,k)
    ------------------------------------------------------
    Entrees:
       int i,j, k      : Numeros axes de rotation (X=1, Y=2, Z=3)
       double *angs    : Adresse des angles ltr (radians)
	   double *m	   : Adresse matrice affectée

    Sorties:
		m est calculée
		Retour		: Adresse de la matrice affectée
----------------------------------------------------------------------*/
double *angs2mrot(double* m, double* angs, int i, int j, int k)
{
double p[9], q[9] ;
    na2mrot(q,i,angs[0]) ;
    na2mrot(m,j,angs[1]) ;
    m3m3(p,q,m) ;
    na2mrot(q,k,angs[2]) ;
    m3m3(m,p,q) ;
    return m;
}

/*----------------------------------------------------------------------
	Extraction des angles d'Euler : précession, nutation,
       rotation propre (/Z0,/Y1,/X2) d'une matrice de rotation
	----------------------------------------------------------

	Entrees:
		double *euler	: Adresse des angles Psi, teta, phi(radians)
		double *A		: Adresse de la matrice de rotation donné

	Sorties:
		angles euler calculés (radians)
		Retour		: Adresse des angles carol

    Remarque :
		Pour abs(cos(teta)) > 1. - MY_EPSREL conservation de la somme ou
		de la difference des valeurs passées en référence qui
		sont également utilisées pour envoyer la solution dans la
		même nappe
----------------------------------------------------------------------*/
double *mrot2euler(double *euler, double *a)
{
double psi, tet, phi, som, dif;

	tet = acos(a[8]);
	if (fabs(a[8]) < 1. - MY_EPSREL)
	{
		psi = atan2(a[6],-a[7]);
		phi = atan2(a[2], a[5]);
	}
	else
	{
		if (a[8] < 0.)
		{
			som = euler[0] + euler[2];
			dif = atan2(a[1], a[0]);
		}
		else
		{
			dif = euler[0] - euler[2];
			som = atan2(a[1], a[0]);
		}
		psi = (som+dif)/2.;
		phi = (som-dif)/2.;
	}
	/* On met dans la même nappe que la solution précédente */
	if (fabs(euler[1]) < 0.)
	{
	   if (psi < 0.) psi =  psi + M_PI; else psi =   psi - M_PI;
	   if (phi < 0.) phi =  phi + M_PI; else phi =   phi - M_PI;
	   tet = -tet ;
	}
	euler[0] = psi;
	euler[1] = tet;
	euler[2] = phi;
	return euler ;
}

/*----------------------------------------------------------------------
	Extraction des angles aero. : Cap, Assiette, Roulis
	                   (/Z0,/Y1,/X2) d'une matrice de rotation
	----------------------------------------------------------

	Entrees:
		double *carol	: Adresse des angles Cap, Ass., Rol. (radians)
		double *A		: Adresse de la matrice de rotation donné

	Sorties:
		angles carol calculés (radians)
		Retour		: Adresse des angles carol

    Remarque :
		Pour abs(ass) > pi/2 - 1.E-6   conservation de la somme ou
		de la difference des valeurs passées en référence qui
		sont également utilisées pour envoyer la solution dans la
		même nappe
----------------------------------------------------------------------*/
double *mrot2carol(double *carol, double *a)
{
double psi, tet, phi, som, dif, pis2 = M_PI/2.;

	tet = -asin(a[2]);
	if (fabs(tet) < pis2 - 1.e-6)
	{
		psi = atan2(a[1],a[0]);
		phi = atan2(a[5],a[8]);
	}
	else
	{
		if (tet > 0.)
		{
			som = carol[0]+carol[2];
			dif = atan2(-a[3],a[4]);
		}
		else
		{
			dif = carol[0]-carol[2];
			som = atan2(-a[3],a[4]);
		}
		psi = (som+dif)/2.;
		phi = (som-dif)/2.;
	}
	/* On met dans la même nappe que la solution précédente */
	if (fabs(carol[1]) > pis2)
	{
	   if (psi < 0.) psi =  psi + M_PI; else psi =   psi - M_PI;
	   if (phi < 0.) phi =  phi + M_PI; else phi =   phi - M_PI;
	   if (tet > 0.) tet = M_PI - tet ; else tet = -M_PI - tet;
	}
	carol[0] = psi;
	carol[1] = tet;
	carol[2] = phi;
	return carol ;
}

/*----------------------------------------------------------------------
	Extraction des angles l,t,r (/Z0,/X1,/Y2) d'une matrice de rotation
	--------------------------------------------------------------------

	Entrees:
		double *ltr	: Adresse des angles l,t,r  (radians)
		double *A	: Adresse de la matrice de rotation donné

	Sorties:
		angles ltr calculés (radians)
		Retour		: Adresse des angles l,t,r

    Remarque :
		Pour abs(t) > pi/2 - 1.E-6   conservation de la somme ou de la
		difference des valeurs ltr passées en référence.
----------------------------------------------------------------------*/
double *mrot2ltr (double *ltr, double *A)
{
double l, t, r, som,dif,pis2 ;

	pis2 = M_PI/2. ;
	t = asin( A[5] ) ;
	if ( fabs(t) < pis2 - 1.E-6)
	{
	   l = atan2( -A[3] , A[4] ) ;
	   r = atan2( -A[2] , A[8] ) ;
	}
	else
	{
	   if ( t > 0.)
	   {
	      som = atan2( A[1],A[0] ) ;
	      dif = ltr[0] - ltr[2] ;
	   }
	   else
	   {
	      dif = atan2( A[1],A[0] ) ;
	      som = ltr[0] + ltr[2] ;
	   }
	   l = (som + dif)/2. ;
	   r = (som - dif)/2. ;
	}

	if(fabs(ltr[1]) > pis2)
	{
	   if(l < 0.) l = M_PI + l ; else l = -M_PI + l ;
	   if(t > 0.) t = M_PI - t ; else t = -M_PI - t ;
	   if(r < 0.) r = M_PI + r ; else r = -M_PI + r ;
	}
	ltr[0] = l ;
	ltr[1] = t ;
	ltr[2] = r ;

	return ltr;
}

#define MY_EPSREL 1.E-14

/*----------------------------------------------------------------------
    CALCUL DES 3 ANGLES DE ROTATION ASSOCIES A UNE MATRICE DE ROTATION
    ------------------------------------------------------------------
Entrees:
  angs : Pointeur sur les 3 angles de rotation (en radians) cherches. Les
         valeurs entrees ne sont utilisees que si la configuration
         angulaire trouvee est singuliere.

  r    : Pointeur sur les elements de la matrice rotation rangee en
         memoire colonne par colonne (rangement compatible Linpack,
         Fortan, Matlab, Scilab, etc...)

  l,m,n: Numeros des axes de base (1, 2 ou 3) autour desquels se font les
         rotations successives. Par exemple dans le cas l=3, m=2, n=1 les
         rotations sont dans l'ordre :
         - la rotation d'angle angs[0] autour de ZO=Z1 (Z est l'axe
         d'index l=3) qui transforme la base B0 (initiale) en base B1,
         - puis la rotation d'angle angs[1] autour de Y1=Y2 (Y est l'axe
         d'index m=2) qui transforme la base B1 en base B2,
         - puis la rotation d'angle angs[2] autour de X2=X3 (X est l'axe
         d'index n=1) qui transforme la base B2 en base B3 (finale).

Sorties:
  angs : Les 3 angles de rotation (en radian) sont calcules.

ATTENTION : En configuration réguliere, la solution est fournie dans la
  premiere nappe, alors que la celle qui est recherchee est peut etre
  dans la deuxieme nappe.

Valeur de retour :
  0 si traité en regulier, 1 si traité en singulier.
----------------------------------------------------------------------*/
int mrot2angs(double* angs, double* r, int l, int m, int n)
{
double var, som, dif;
int flagsing, M, p;

    --l; --m ; --n; /* On ramene les indices 1,2,3 a 0,1,2 */

    if (n != l) /* Cas Cardan (rotation autour des 3 axes) */
    {
        var = r[l+n*3];
        if (n == l+1 || (l==2 && n==0)) /* Cas 132, 213, 321 */
        {
            if (fabs(var) < 1. - MY_EPSREL) /* regulier */
            {
                flagsing = 0;
	            angs[1] = -asin(var);
	            angs[0] = atan2(r[m+n*3],r[n+n*3]);
	            angs[2] = atan2(r[l+m*3],r[l+l*3]);
            } 
            else /* var = +/- 1 : singulier */
            {
                flagsing = 1;
                if(var > 0.) 
                {
                    if (var > 1.) angs[1] = -M_PI/2;
                    else angs[1] = -asin(var);
                    som = atan2(-r[n+m*3],r[m+m*3]);
                    dif = angs[0] - angs[2];
                }
                else
                {
                    if (var < -1.) angs[1] =  M_PI/2;
                    else angs[1] = -asin(var);
                    dif = atan2(-r[n+m*3],r[m+m*3]);
                    som = angs[0] + angs[2];
                }
            }
	    }
	    else /* Cas 123, 231, 312 */
	    {
            if (fabs(var) < 1. - MY_EPSREL) /* regulier */
            {
                flagsing = 0;
	            angs[1] = asin(var);
	            angs[0] = -atan2(r[m+n*3],r[n+n*3]);
	            angs[2] = -atan2(r[l+m*3],r[l+l*3]);
            }
            else /* var = +/- 1 : singulier */
            {
                flagsing = 1;
                if(var > 0.) 
                {
                    if (var > 1.) angs[1] = M_PI/2;
                    else angs[1] = asin(var);
                    som = atan2(r[n+m*3],r[m+m*3]);
                    dif = angs[0] - angs[2];                
                }
                else
                {
                    if (var < -1.) angs[1] = -M_PI/2;
                    else angs[1] = asin(var);
                    dif = atan2(r[n+m*3],r[m+m*3]);
                    som = angs[0] + angs[2];
                }
            }
	    }
    }
    else /* Cas Euler (un des 3 axes n'est pas utilse) */
    {	  
        var = r[l+n*3];
        M = m + 1; if (M == 3) M = 0;
        if (n == M) /* Cas 131, 212, 323 */
        {
            p = l + 1; if (p == 3) p = 0;
            if (fabs(var) < 1. - MY_EPSREL) /* regulier */
            {
                flagsing = 0;
                angs[1] = acos(var);
	            angs[0] = atan2(r[m+n*3], r[p+n*3]);
	            angs[2] = atan2(r[l+m*3],-r[l+p*3]);
            }
            else /* var = +/- 1 : singulier */
            {
                flagsing = 1;
                if (var > 0.)
                {
                    if (var > 1.) angs[1] = 0.;
                    else angs[1] = acos(var);
                    som =atan2(-r[p+m*3],r[m+m*3]);
                    dif = angs[0] - angs[2];
                }
                else
                {
                    if (var < -1.) angs[1] = M_PI;
                    else angs[1] = acos(var);
                    dif = atan2(-r[p+m*3],r[m+m*3]);
                    som = angs[0] + angs[2];
               }
            }
	    }
        else /* Cas 121, 232, 313 */
        {
            p = M;
            if (fabs(var) < 1. - MY_EPSREL) /* regulier */
            {
                flagsing = 0;
                angs[1] = acos(var);
	            angs[0] = atan2(r[m+n*3],-r[p+n*3]);
	            angs[2] = atan2(r[l+m*3], r[l+p*3]);
            }
            else /* var = +/- 1 : singulier */
            {
                flagsing = 1;
                if (var > 0.)
                {
                    if (var > 1.) angs[1] = 0.;
                    else angs[1] = acos(var);
                    som =atan2( r[p+m*3],r[m+m*3]);
                    dif = angs[0] - angs[2];
                }
                else
                {
                    if (var < -1.) angs[1] = M_PI;
                    else angs[1] = acos(var);
                    dif = atan2( r[p+m*3],r[m+m*3]);
                    som = angs[0] + angs[2];
                }
            }
	    }
    }
    if(flagsing)
    {
		angs[0] = (som+dif)/2.;
		angs[2] = (som-dif)/2.;
    }
    return flagsing ;
}

/*----------------------------------------------------------------------
          MODIFICATION ANGLES POUR CONTINUITE DE SOLUTION
          -----------------------------------------------
Entrees:
  angs : Pointeur sur les 3 angles (radians) de rotation a eventuellement
         modifier pour qu'ils y ait une continuite avec les 3 angles
         angsref

  angsref : Pointeur sur les 3 angles (radians) de reference

  l,m,n: Numeros des axes de base (1, 2 ou 3) autour desquels se font les
         rotations successives. Par exemple dans le cas l=3, m=2, n=1 les
         rotations sont dans l'ordre :
         - la rotation d'angle angs[0] autour de ZO=Z1 (Z est l'axe
         d'index l=3) qui transforme la base B0 (initiale) en base B1,
         - puis la rotation d'angle angs[1] autour de Y1=Y2 (Y est l'axe
         d'index m=2) qui transforme la base B1 en base B2,
         - puis la rotation d'angle angs[2] autour de X2=X3 (X est l'axe
         d'index n=1) qui transforme la base B2 en base B3 (finale).

Sorties:
  angs : Les 3 angles de rotation (en radian) sont eventuellement
         modifies
----------------------------------------------------------------------*/
void angscont(double *angs, double* angsref, int l, int m, int n)
{
double dif;
int i;

    dif = angs[0] - angsref[0];
    while (dif >  M_PI) dif -= 2*M_PI;
    while (dif < -M_PI) dif += 2*M_PI;
    if (fabs(dif) > M_PI/2) /* nappes differentes */
    {
        angs[0] += M_PI;
        angs[2] += M_PI;
        if (n != l) angs[1] = M_PI - angs[1];
        else        angs[1] =       -angs[1];
    }
    for (i = 0; i < 3; i++)
    {
        while (angs[i]-angsref[i] >  M_PI) angs[i] -= 2*M_PI;
        while (angs[i]-angsref[i] < -M_PI) angs[i] += 2*M_PI;
    }
}

/*---------------------------------------------------------------------
	Génération matrice de rotation d'angle a autour de Ox
	-----------------------------------------------------

	Entrees:
		double *m	: Adresse de la matrice affectée
		double a	: Angle de rotation (radians)

	Sorties:
		m est affectée
		Retour		: Adresse de la matrice affectée
----------------------------------------------------------------------*/
double *mrotx(double *m, double a)
{
double s = sin(a), c = cos(a);
	m[0] = 1.; m[3] = 0.; m[6] = 0.;
	m[1] = 0.; m[4] = c ; m[7] =-s ;
	m[2] = 0.; m[5] = s ; m[8] = c ;
    return m ;
}

/*---------------------------------------------------------------------
	Génération matrice de rotation d'angle a autour de Oz
	-----------------------------------------------------

	Entrees:
		double *m	: Adresse de la matrice affectée
		double a	: Angle de rotation  (radians)

	Sorties:
		m est affectée
		Retour		: Adresse de la matrice affectée
----------------------------------------------------------------------*/
double *mrotz(double *m, double a)
{
double s = sin(a), c = cos(a);
	m[0] = c ; m[3] =-s ; m[6] = 0.;
	m[1] = s ; m[4] = c ; m[7] = 0.;
	m[2] = 0.; m[5] = 0.; m[8] = 1.;
    return m ;
}

/*---------------------------------------------------------------------
	Génération matrice de rotation d'angle a autour de Oy
	-----------------------------------------------------

	Entrees:
		double *m	: Adresse de la matrice affectée
		double a	: Angle de rotation  (radians)

	Sorties:
		m est affectée
		Retour		: Adresse de la matrice affectée
----------------------------------------------------------------------*/
double *mroty(double *m, double a)
{
double s = sin(a), c = cos(a);
	m[0] = c ; m[3] = 0.; m[6] = s ;
	m[1] = 0.; m[4] = 1.; m[7] = 0.;
	m[2] =-s ; m[5] = 0.; m[8] = c ;
    return m ;
}
/*---------------------------------------------------------------------
	Génération matrice de rotation d'angle a autour du vecteur V
	------------------------------------------------------------

	Entrees:
		double *m	: Adresse de la matrice affectée
		double a	: Angle de rotation  (radians)
		double *v	: Adresse du vecteur unitaire porté
		              par l'axe de rotation

	Sorties:
		m est affectée
		Retour		: Adresse de la matrice affectée
----------------------------------------------------------------------*/
double *av3tomrot(double* m, double a, double* v)
{
    double wx, wy, wz;
    double k;

    wx = v[0];
    wy = v[1];
    wz = v[2];
    k = 1.-cos(a);
    m[0] = 1. + k * (wx * wx - 1.);
    m[4] = 1. + k * (wy * wy - 1.);
    m[8] = 1. + k * (wz * wz - 1.);
    m[1] = k * wx * wy;
    m[2] = k * wz * wx;
    m[5] = k * wy * wz;
    k = sin(a);
    wx = k * wx;
    wy = k * wy;
    wz = k * wz;
    m[7] = m[5] - wx;
    m[5] += wx;
    m[6] = m[2] + wy;
    m[2] -= wy;
    m[3] = m[1] - wz;
    m[1] += wz;
	return m;
}

/*---------------------------------------------------------------------
	Génération matrice de rotation autour du vecteur , d'angle égal
	                         à la longeur de V
	---------------------------------------------------------------

	Entrees:
		double *m	: Adresse de la matrice affectée
		double *v	: Adresse du vecteur rotation

	Sorties:
		m est affectée
		Retour		: Adresse de la matrice affectée
----------------------------------------------------------------------*/
double *v3tomrot(double* m, double* v)
{
double a, u[3];

	a = lunitv3(u,v) ;
	return av3tomrot(m, a, u);
}


/*----------------------------------------------------------------------
	Extraction de l'angle de rotation d'une matrice de rotation
	--------------------------------------------------------------------

	Entree:
		double *A	: Adresse de la matrice de rotation donné

	Sortie:
		Retour		: angle de rotation en radian
----------------------------------------------------------------------*/
double mrot2ang(double* a)
{
	return acos(0.5 *(a[0] + a[4] + a[8]-1.)) ;
}


/*----------------------------------------------------------------------
	Extraction des paramètres d'euler d'une matrice de rotation,
	à savoir la direction (vecteur unitaire) de l'axe de rotation
	unique et de l'angle de rotation correspondante.
	--------------------------------------------------------------------

	Entrees:
		double *v	: Adresse du vecteur
		double *A	: Adresse de la matrice de rotation donné

	Sorties:
		V unité est calculé
		Retour		: angle de rotation en radian

    Remarque : Solution particulière pour angle voisin de Pi.
----------------------------------------------------------------------*/
double mrot2av3(double* v, double* a)
{
double t, s, ss, cs ;
int k ;

    t = a[0] + a[4] + a[8];
	v[0] = a[5] - a[7];
	v[1] = a[6] - a[2];
	v[2] = a[1] - a[3];
	ss = 0.5 * lunitv3(v, v);
	cs = 0.5 * (t - 1.);
    if (cs > 0. || ss > 1.e-6) return (atan2(ss, cs)) ;
	else
	{
	/* Cas particulier angle voisin de Pi */
		s = acos(cs) ;
	    t = 3. - t;
	    k = 0;
	    if (a[4] > a[0] && a[4] > a[8]) k = 1;
	    if (a[8] > a[0] && a[8] > a[4]) k = 2;
	    switch (k)
	    {
	    case 0:
			v[0] = sqrt((t - 2. + 2 * a[0])/t);
			if (a[7] > a[5]) v[0] = -v[0];
			t *= v[0];
			v[1] = (a[3] + a[1])/t;
			v[2] = (a[6] + a[2])/t;
			break ;
	    case 1:
			v[1] = sqrt((t - 2. + 2 * a[4])/t);
			if (a[2] > a[6]) v[1] = -v[1];
			t *= v[1];
			v[2] = (a[5] + a[7])/t;
			v[0] = (a[1] + a[3])/t;
			break ;
	    case 2:
			v[2] = sqrt((t - 2. + 2 * a[8])/t);
			if (a[3] > a[1]) v[2] = -v[2];
			t *= v[2];
			v[0] = (a[6] + a[2])/t;
			v[1] = (a[7] + a[5])/t;
			break ;
	    }
	}
    return s ;
}

/*----------------------------------------------------------------------
	Extraction du vecteur rotation d'une matrice de rotation,
	à savoir un vecteur de direction l'axe de rotation et de longueur
	l'angle de la rotation correspondante.
	--------------------------------------------------------------------

	Entrees:
		double *v	: Adresse du vecteur
		double *A	: Adresse de la matrice de rotation donné

	Sorties:
		V est calculé
		Retour		: Adresse de V
----------------------------------------------------------------------*/
double *mrot2v3(double* v, double* a)
{
	double s = mrot2av3(v, a) ;
	return scav3(v, s, v);
}

/*---------------------------------------------------------------------
	Composantes Omega dans Rfixe en fonction de la dérivée des
									angles d'euler
	----------------------------------------------------------

	Entrees:
		double *Omo		: Adresse vecteur Omega
		double *eulerp	: Adresse dérivées des euler
		double *euler		: Adresse angles euler

	Sorties:
		p,q,r dans Ro calculés
		Retour		: Adresse de Omo
----------------------------------------------------------------------*/
double *deuler2omo (double *omo, double *eulerp, double *euler)
{
double cp, sp, ct, st ;

    cp = cos (euler[0]) ;
    sp = sin (euler[0]) ;
    ct = cos (euler[1]) ;
    st = sin (euler[1]) ;
    omo[0] =           eulerp[1] * cp + eulerp[2] * sp * st ;
    omo[1] =           eulerp[1] * sp - eulerp[2] * cp * st ;
    omo[2] = eulerp[0] +                eulerp[2] * ct ;
	return omo;
}

/*---------------------------------------------------------------------
	Composantes Omega dans Rmobile en fonction de la dérivée des
									angles d'euler
	----------------------------------------------

	Entrees:
		double *Omm		: Adresse vecteur Omega
		double *eulerp	: Adresse dérivées des euler
		double *euler		: Adresse angles euler

	Sorties:
		p,q,r dans Rm calculés
		Retour		: Adresse de Omm
----------------------------------------------------------------------*/
double *deuler2omm (double *omm, double *eulerp, double *euler)
{
double ct, st, cf, sf ;

    ct = cos (euler[1]) ;
    st = sin (euler[1]) ;
    cf = cos (euler[2]) ;
    sf = sin (euler[2]) ;
    omm[0] = - eulerp[0] * st * sf + eulerp[1] * cf ;
    omm[1] =   eulerp[0] * st * sf - eulerp[1] * sf ;
    omm[2] =   eulerp[0] * ct                       + eulerp[2] ;
	return omm;
}

/*---------------------------------------------------------------------
	Composantes Omega dans Rfixe en fonction de la dérivée des
									angles aéro.
	----------------------------------------------------------

	Entrees:
		double *Omo		: Adresse vecteur Omega
		double *carolp	: Adresse dérivées des angles aéro.
		double *carol		: Adresse angles aéro.

	Sorties:
		p,q,r dans Ro calculés
		Retour		: Adresse de Omo
----------------------------------------------------------------------*/
double *dcarol2omo (double *omo, double *carolp, double *carol)
{
double cp, sp, ct, st ;

    cp = cos (carol[0]) ;
    sp = sin (carol[0]) ;
    ct = cos (carol[1]) ;
    st = sin (carol[1]) ;
    omo[0] =          - carolp[1] * sp + carolp[2] * cp * ct ;
    omo[1] =            carolp[1] * cp + carolp[2] * sp * ct ;
    omo[2] = carolp[0]                 - carolp[2]      * st ;
	return omo;
}

/*---------------------------------------------------------------------
	Composantes Omega dans Rmobile en fonction de la dérivée des
									angles aéro.
	----------------------------------------------

	Entrees:
		double *Omm		: Adresse vecteur Omega
		double *carolp	: Adresse dérivées des angles aéro.
		double *carol		: Adresse angles aéro.

	Sorties:
		p,q,r dans Rm calculés
		Retour		: Adresse de Omm
----------------------------------------------------------------------*/
double *dcarol2omm (double *omm, double *carolp, double *carol)
{
double ct, st, cf, sf ;

    ct = cos (carol[1]) ;
    st = sin (carol[1]) ;
    cf = cos (carol[2]) ;
    sf = sin (carol[2]) ;
    omm[0] = - carolp[0]      * sf                  + carolp[2] ;
    omm[1] =   carolp[0] * ct * sf + carolp[1] * cf ;
    omm[2] =   carolp[0] * ct * cf - carolp[1] * sf ;
	return omm;
}

/*---------------------------------------------------------------------
	Composantes Omega dans Rfixe en fonction de la dérivée des
									angles ltr
	------------------------------------------

	Entrees:
		double *Omo		: Adresse vecteur Omega
		double *ltrp	: Adresse dérivées des ltr
		double *ltr		: Adresse angles ltr

	Sorties:
		p,q,r dans Ro calculés
		Retour		: Adresse de Omo
----------------------------------------------------------------------*/
double *dltr2omo (double *omo, double *ltrp, double *ltr)
{
double ct, st, cl, sl ;

    ct = cos (ltr[1]) ;
    st = sin (ltr[1]) ;
    cl = cos (ltr[0]) ;
    sl = sin (ltr[0]) ;
    omo[0] =           ltrp[1] * cl - ltrp[2] * sl * ct ;
    omo[1] =           ltrp[1] * sl + ltrp[2] * cl * ct ;
    omo[2] = ltrp[0] +                ltrp[2] * st ;
	return omo;
}

/*---------------------------------------------------------------------
	Composantes Omega dans Rmobile en fonction de la dérivée des
									angles ltr
	------------------------------------------

	Entrees:
		double *Omm		: Adresse vecteur Omega
		double *ltrp	: Adresse dérivées des ltr
		double *ltr		: Adresse angles ltr

	Sorties:
		p,q,r dans Rm calculés
		Retour		: Adresse de Omm
----------------------------------------------------------------------*/
double *dltr2omm (double *omm, double *ltrp, double *ltr)
{
double ct, st, cr, sr ;

    ct = cos (ltr[1]) ;
    st = sin (ltr[1]) ;
    cr = cos (ltr[2]) ;
    sr = sin (ltr[2]) ;
    omm[0] = - ltrp[0] * sr * ct + ltrp[1] * cr ;
    omm[1] =   ltrp[0] *      st                + ltrp[2] ;
    omm[2] =   ltrp[0] * cr * ct + ltrp[1] * sr ;
	return omm;
}

/*---------------------------------------------------------------------
	Dérivées des angles d'euler en fonction de Omega dans Ro
	--------------------------------------------------------

	Entrees:
		double *eulerp	: Adresse dérivées des angles d'euler
		double *Omo		: Adresse vecteur Omega dans Ro
		double *euler	: Adresse angles d'euler

	Sorties:
		dérivées calculées
		Retour		: Adresse des dérivées
----------------------------------------------------------------------*/
double *omo2deuler (double *eulerp, double *omo, double *euler)
{
double ct, st, cp, sp, bid, som, dif ;

    ct = cos (euler[1]) ;
    st = sin (euler[1]) ;
    cp = cos (euler[0]) ;
    sp = sin (euler[0]) ;

	eulerp[1] = cp*omo[0] + sp*omo[1] ;

	bid = sp*omo[0] - cp*omo[1] ;
	if (fabs(st) > 1.e-6)
	{
		eulerp[2] =  bid/st ;
		eulerp[0] =  omo[2] - ct*eulerp[2] ;
	}
	else
	{
		if (ct > 0.) {som = omo[2] ; dif = eulerp[0]-eulerp[2] ;}
		else         {som = eulerp[0]+eulerp[2] ; dif = omo[2] ;}
		eulerp[0] = 0.5*(som + dif) ;
		eulerp[2] = 0.5*(som - dif) ;
	}
	return eulerp ;
}

/*---------------------------------------------------------------------
	Dérivées des angles d'euler en fonction de Omega dans Rm
	--------------------------------------------------------

	Entrees:
		double *eulerp	: Adresse dérivées des angles d'euler
		double *Omo		: Adresse vecteur Omega dans Rm
		double *euler		: Adresse angles d'euler

	Sorties:
		dérivées calculées
		Retour		: Adresse des dérivées
----------------------------------------------------------------------*/
double *omm2deuler (double *eulerp, double *omm, double *euler)
{
double ct, st, cf, sf, bid, som, dif ;

    ct = cos (euler[1]) ;
    st = sin (euler[1]) ;
    cf = cos (euler[2]) ;
    sf = sin (euler[2]) ;

	eulerp[1] =  cf*omm[0] - sf*omm[1] ;

	bid = sf*omm[0] + cf*omm[1] ;
	if (fabs(ct) > 1.e-6)
	{
		eulerp[0] =  bid/st ;
		eulerp[2] =  omm[2] - ct*eulerp[0] ;
	}
	else
	{
		if (ct > 0.) {som = omm[2] ; dif = eulerp[2]-eulerp[0] ;}
		else         {som = eulerp[0]+eulerp[2] ; dif = omm[2] ;}
		eulerp[2] = 0.5*(som + dif) ;
		eulerp[0] = 0.5*(som - dif) ;
	}
	return eulerp ;
}

/*---------------------------------------------------------------------
	Dérivées des angles aero en fonction de Omega dans Ro
	----------------------------------------------------

	Entrees:
		double *carolp	: Adresse dérivées des angles aero
		double *Omo		: Adresse vecteur Omega dans Ro
		double *carol	: Adresse angles aero

	Sorties:
		dérivées calculées
		Retour		: Adresse des dérivées
----------------------------------------------------------------------*/
double *omo2dcarol (double *carolp, double *omo, double *carol)
{
double ct, st, cp, sp, bid, som, dif ;

    ct = cos (carol[1]) ;
    st = sin (carol[1]) ;
    cp = cos (carol[0]) ;
    sp = sin (carol[0]) ;

	carolp[1] = - sp*omo[0] + cp*omo[1] ;

	bid = cp*omo[0] + sp*omo[1] ;
	if (fabs(ct) > 1.e-6)
	{
		carolp[2] =  bid/ct ;
		carolp[0] =  omo[2] + st*carolp[2] ;
	}
	else
	{
		if (st < 0.) {som = omo[2] ; dif = carolp[0]-carolp[2] ;}
		else         {som = carolp[0]+carolp[2] ; dif = omo[2] ;}
		carolp[0] = 0.5*(som + dif) ;
		carolp[2] = 0.5*(som - dif) ;
	}
	return carolp ;
}

/*---------------------------------------------------------------------
	Dérivées des angles aero en fonction de Omega dans Rm
	----------------------------------------------------

	Entrees:
		double *carolp	: Adresse dérivées des angles aero
		double *Omo		: Adresse vecteur Omega dans Rm
		double *carol		: Adresse angles aero

	Sorties:
		dérivées calculées
		Retour		: Adresse des dérivées
----------------------------------------------------------------------*/
double *omm2dcarol (double *carolp, double *omm, double *carol)
{
double ct, st, cf, sf, bid, som, dif ;

    ct = cos (carol[1]) ;
    st = sin (carol[1]) ;
    cf = cos (carol[2]) ;
    sf = sin (carol[2]) ;

	carolp[1] =  cf*omm[1] - sf*omm[2] ;

	bid = sf*omm[1] + cf*omm[2] ;
	if (fabs(ct) > 1.e-6)
	{
		carolp[0] =  bid/ct ;
		carolp[2] =  omm[0] + st*carolp[0] ;
	}
	else
	{
		if (st < 0.) {som = omm[0] ; dif = carolp[2]-carolp[0] ;}
		else         {som = carolp[0]+carolp[2] ; dif = omm[0] ;}
		carolp[2] = 0.5*(som + dif) ;
		carolp[0] = 0.5*(som - dif) ;
	}
	return carolp ;
}

/*---------------------------------------------------------------------
	Dérivées des angles ltr en fonction de Omega dans Ro
	----------------------------------------------------

	Entrees:
		double *ltrp	: Adresse dérivées des ltr
		double *Omo		: Adresse vecteur Omega dans Ro
		double *ltr		: Adresse angles ltr

	Sorties:
		dérivées calculées
		Retour		: Adresse des dérivées
----------------------------------------------------------------------*/
double *omo2dltr (double *ltrp, double *omo, double *ltr)
{
double ct, st, cl, sl, bid, som, dif ;

    ct = cos (ltr[1]) ;
    st = sin (ltr[1]) ;
    cl = cos (ltr[0]) ;
    sl = sin (ltr[0]) ;

	ltrp[1] =  cl*omo[0] + sl*omo[1] ;

	bid = cl*omo[1] - sl*omo[0] ;
	if (fabs(ct) > 1.e-6)
	{
		ltrp[2] =  bid/ct ;
		ltrp[0] =  omo[2] - st*ltrp[2] ;
	}
	else
	{
		if (st > 0.) {som = omo[2] ; dif = ltrp[0]-ltrp[2] ;}
		else         {som = ltrp[0]+ltrp[2] ; dif = omo[2] ;}
		ltrp[0] = 0.5*(som + dif) ;
		ltrp[2] = 0.5*(som - dif) ;
	}
	return ltrp ;
}

/*---------------------------------------------------------------------
	Dérivées des angles ltr en fonction de Omega dans Rm
	----------------------------------------------------

	Entrees:
		double *ltrp	: Adresse dérivées des ltr
		double *Omo		: Adresse vecteur Omega dans Rm
		double *ltr		: Adresse angles ltr

	Sorties:
		dérivées calculées
		Retour		: Adresse des dérivées
----------------------------------------------------------------------*/
double *omm2dltr (double *ltrp, double *omm, double *ltr)
{
double ct, st, cr, sr, bid, som, dif ;

    ct = cos (ltr[1]) ;
    st = sin (ltr[1]) ;
    cr = cos (ltr[2]) ;
    sr = sin (ltr[2]) ;

	ltrp[1] =  cr*omm[0] + sr*omm[2] ;

	bid = cr*omm[2] - sr*omm[0] ;
	if (fabs(ct) > 1.e-6)
	{
		ltrp[0] =  bid/ct ;
		ltrp[2] =  omm[1] - st*ltrp[0] ;
	}
	else
	{
		if (st > 0.) {som = omm[1] ; dif = ltrp[2]-ltrp[0] ;}
		else         {som = ltrp[0]+ltrp[2] ; dif = omm[1] ;}
		ltrp[2] = 0.5*(som + dif) ;
		ltrp[0] = 0.5*(som - dif) ;
	}
	return ltrp ;
}

/*---------------------------------------------------------------------
	Impression d'un Por de R3
	------------------------------

	Entrees:
		char *s		: Texte à impimer (RC ajouté en fin du texte)
		double *A	: Adresse du Por donnée à imprimer

    Remarque : Translation en première colonne.
----------------------------------------------------------------------*/
void printpor(char *s, double *a)
{
int i, j;

	sprintf(strfmt+1,"%d.%dl%c",(printformat/10)%100,(printformat%10),
		printformat/1000 ? 'g' : 'f');

	printf("%s\n",s);
	for (i = 0 ; i < 3 ; i++)
	{
		printf("%10.3f | ",a[i]) ;
		for(j=i+3 ; j<12 ; j += 3) printf(strfmt,a[j]) ;
		printf("\n");
	}
}

/*---------------------------------------------------------------------
	Recopie d'un Por : U = V
	------------------------

	Entrees:
		double *U	: Adresse du por affecté
		double *V	: Adresse du por donnée

	Sorties:
		U est affecté
		Retour		: Adresse du por affecté
----------------------------------------------------------------------*/
double *cpypor(double* u, double* v)
{
    u[0] = v[0] ;
    u[1] = v[1] ;
    u[2] = v[2] ;
    u[3] = v[3] ;
    u[4] = v[4] ;
    u[5] = v[5] ;
    u[6] = v[6] ;
    u[7] = v[7] ;
    u[8] = v[8] ;
    u[9] = v[9] ;
    u[10] = v[10] ;
    u[11] = v[11] ;
	return u ;
}

/*---------------------------------------------------------------------
	Produit Por x Por de R3 : A = B C
	---------------------------------

	Entrees:
		double *A	: Adresse du Por affecté
		double *B	: Adresse du premier Por donné
		double *C	: Adresse du deuxième Por donné

	Sorties:
		A est affecté
		Retour		: Adresse du Por affecté

    Remarque : A peut occuper la même mémoire que B ou C
----------------------------------------------------------------------*/
double *porpor(double* a, double* b, double* c)
{
double d[3] ;
    m3v3(d, b+3, c);
    m3m3(a+3, b+3, c+3);
    addv3(a, b, d) ;
	return a;
}

/*---------------------------------------------------------------------
	Produit Por x Por inverse de R3 : A = B inv(C)
	----------------------------------------------

	Entrees:
		double *A	: Adresse du Por affecté
		double *B	: Adresse du premier Por donné
		double *C	: Adresse du deuxième Por donné

	Sorties:
		A est affecté
		Retour		: Adresse du Por affecté

    Remarque : A peut occuper la même mémoire que B ou C
----------------------------------------------------------------------*/
double *porpori(double* a, double* b, double* c)
{
double d[3] ;
    m3tm3(a+3, b+3 ,c+3) ;
    m3v3(d, a+3, c) ;
    subv3(a, b, d) ;
	return a;
}

/*---------------------------------------------------------------------
	Produit Por inverse x Por de R3 : A = inv(B) C
	----------------------------------------------

	Entrees:
		double *A	: Adresse du Por affecté
		double *B	: Adresse du premier Por donné
		double *C	: Adresse du deuxième Por donné

	Sorties:
		A est affecté
		Retour		: Adresse du Por affecté

    Remarque : A peut occuper la même mémoire que B ou C
----------------------------------------------------------------------*/
double *poripor(double* a, double* b, double* c)
{
    subv3 (a, c, b) ;
    tm3v3 (a, b+3, a) ;
    tm3m3 (a+3, b+3, c+3) ;
	return a;
}
/*---------------------------------------------------------------------
	Produit Por x vecteur de R3 : a = B c
	-------------------------------------

	Entrees:
		double *a	: Adresse vecteur affecté
		double *b	: Adresse du Por donné
		double *c	: Adresse du vecteur donné

	Sorties:
		a est affecté
		Retour		: Adresse vecteur affecté

    Remarque : a peut occuper la même mémoire que c
----------------------------------------------------------------------*/
double *porv3(double* a, double* b,double* c)
{
double d[3] ;
    m3v3(d, b+3, c);
    addv3(a, b, d) ;
	return a;
}

/*---------------------------------------------------------------------
	Produit Por inverse x vecteur de R3 : a = inv(B) c
	--------------------------------------------------

	Entrees:
		double *a	: Adresse vecteur affecté
		double *b	: Adresse du Por donné
		double *c	: Adresse du vecteur donné

	Sorties:
		a est affecté
		Retour		: Adresse vecteur affecté

    Remarque : a peut occuper la même mémoire que c
----------------------------------------------------------------------*/
double *poriv3(double* a, double* b,double* c)
{
    subv3 (a, c, b) ;
    tm3v3 (a, b+3, a) ;
	return a;
}

