#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "julien.h"

#define GLOBAL 

/** Cte : nombre de secondes par jour */
GLOBAL const double Date_jour2s = 86400.;

/** Cte : valeur de J2000 (01/01/2000 a 12h) en jours julien */
GLOBAL const double Date_J2000_j = 2451545.;

/** Cte : Duree J2000 - 1/1/1970 0h en secondes */
GLOBAL const double Date_Sys2J2000_s = 946728000.;

/** Cte : valeur de J1900 (31/12/1899 a 12h) en jours julien */
GLOBAL const double Date_J1900_j = 2415020.;

/*
        Dans les formules de cinématique spatiale on utilise le temps des éphémérides TE (actuellement on dit le temps terrestre TT).
        C'est un temps linéaire uniforme qui suit le temps atomique international TAI à 32.184 seconde :
                                     TE = TAI + 32.184
        Notre temps civil (UTC) utilisé dans notre vie sociale et économique est également uniforme linéaire (parallèle au TAI) mais
        varie par paliers d'une seconde pour suivre à moins de 0.9 seconde près le temps physique UT1 directement lié à la rotation
        irrégulière de la terre (mesuré par les géophysiciens) :
         						 	|UTC - UT1| < 0.9
         							 UTC = TAI - n (nombre entier des secondes)
        Le nombre "n" n'est modifié qu'à minuit au passage 30/06 -> 01/07 ou 31/12 -> 01/01, généralement par ajout de la seconde
        numéro 60 à 23h23mn59s.
        
        Pour utiliser les formules de mécanique spaciale, il faut convertir les dates civiles spécifiées en UTC en temps des éphémériques
        en ajoutant deltaET :
        							TE = TAI + 32.184 
        							   = UTC + n + 32.184
        							   = UTC + DeltaET
         mais en interpolant afin de redonner la continuité à TE, alors que UTC présente, lors des variations de n, des marches verticales
         d'une seconde.
 	 */

/**
     * Calcul ecart deltaET entre temps des ephemerides et temps universel
	 * @param date 	date actuelle en jours julien
     * @return	l'ecart deltaET en jours  (TE - TU)
	 */
double jul2deltaET(double jul)
{
    double deltaT, u;
    if (jul < 2328824.) //1664
    {u = (jul-2314580)/36524.219; deltaT = 40.65 + 35.0*u*u;}
    else if (jul < 2396028.) // 1848
    {u = (jul-2451545.)/36523.913; deltaT = 102.12 + u*(102*u + 25.3*u);}
    else if (jul < 2407716.) //1880
    {u = (jul-2396759.)/36525.; deltaT = 6.6 + u*(46.3 + u*(18.8*u - 358.4));}
    else if (jul < 2451545.) //2000
    {u = (jul-2407716.)/365.2417; deltaT = -11.26 + 0.62575*u;}
    else
    {u = (jul-2451545.)/365.2422; deltaT = 63.86 + 0.266875*u;}
    return deltaT/86400.;
}


/** Fonction convertissant un string date du type "2000:05:20 14:00:00" en jour julien
         @param sdate   Entree : le string date
         @return     La date en jour julien
 */
double str19tojul(char *sdate)
{
    int date[6];
    char *ptr;

    date[0] = strtol(sdate, &ptr, 10); if (!*ptr) return 0.;
    date[1] = strtol(ptr+1, &ptr, 10); if (!*ptr) return 0.;
    date[2] = strtol(ptr+1, &ptr, 10); if (!*ptr) return 0.;
    date[3] = strtol(ptr  , &ptr, 10); if (!*ptr) return 0.;
    date[4] = strtol(ptr+1, &ptr, 10); if (!*ptr) return 0.;
    date[5] = strtol(ptr+1, &ptr, 10);
    return tdat2jul(date);

}

/** Fonction ajoutant à la date d'un string date du type "2000:05:20 14:00:00"
    un écart specifie en jour decimaux
         @param sdate   Entree : le string date
         @param ecart   Entree : l'ecart en jour decimaux
         @return     La date en jour julien (pointeur sur l'entree)
*/
char *add2date(char *sdate, double ecart)
{
    char temp[30];
    int date[6];

    jul2tdat(date, str19tojul(sdate) + ecart);
    sprintf(temp,"%04d:%02d:%02d %02d:%02d:%02d",date[0],date[1],date[2],date[3],date[4],date[5]);
    if(strlen(temp)!=19) return NULL;
    strcpy(sdate, temp);
    return sdate;
}

/** Fonction convertissant les dates en jour julien
         @param j    Entree : jour décimal de 1.000 à 31.999 (1.5 : Le 1er à midi)
         @param m    Entree : mois de 1 à 12
         @param a	 Entree : année (attention 1 avant JC est noté 0, etc..)
         @return     La date en jour julien
 */
double dat2jul(double j, int m, int a)
{
    double JJ;
    long M, A;
    /* Changement origine des annees pour qu'elles soient > 0 afin que
            les troncatures soient toutes dans le meme sens */
    A = a + 4800;
    /* Changement origine au 1er Mars, afin que la succession des jours
            soit toujours la meme */
    M = m; if (M <= 2) {A--; M += 12;}
    /* Correction Julienne/gregorienne */
    if (a < 1582 || (a == 1582 && m < 10) || (a == 1582 && m == 10 && j < 15)) JJ = -38.;
    else JJ = (A/400) - (A/100);
    /* Décompte avec troncature magique pour les 30/31 et correction des
            changements d'origine */
    JJ += j + (36525*A/100) + (306*(M+1)/10) - 32167.5;
    return JJ;
}

static char tabjours[7][9] =
{"Lundi", "Mardi", "Mercredi", "Jeudi","Vendredi", "Samedi", "Dimanche"};

/** Fonction donnant le jour de la semaine correspodant à un jour julien donné
         @param JJ    Entree :  date en jour julien
         @return 	  Le jour (Lundi, ...,Dimanche) de la semaine
         */
const char  * jul2jour(double JJ)
{
    long   jj = (long) (JJ + 0.5);
    return tabjours[(int) (jj % 7)];
}


/** Fonction convertissant les jours juliens en date decimale
         @param jdec  Sortie : jdec[0] : jour décimal (1.5 = le 1er à midi)
         @param m	  Sortie : m[0] : mois (1 à 12)
         @param a	  Sortie : a[0] : année (attention 1 avant JC est noté 0, etc..)
         @param JJ    Entree :  date en jour julien
         @return 	  Le jour (Lundi, ...,Dimanche) de la semaine
         */
const char  * jul2dat(double *jdec, int *m, int *a, double JJ)
{
    long   jj, jjm, jrs, mp1, ans, sic;
    /* Decompte les siecles depuis le 1.5/3/-4801 en gregorien etendu
               c'est-à-dire le le 7.5/4/-4801 en calendrier julien */
    jj = (long) (JJ + 0.5);
    sic = (100*jj + 3204500)/3652425;
    /* correction bissextiles julienne/gregorienne */
    jjm = jj + 1486;
    if (jj > 2299160.5) jjm += sic - sic/4; else jjm += 38;
    /* troncatures magiques prenant en compte l'origine au 1er mars */
    ans = (100*jjm - 12210)/36525;
    jrs = (36525*ans)/100;
    mp1 = ((jjm - jrs)*100)/3061;
    jdec[0] = jjm - jrs - ((306*mp1)/10) + JJ + 0.5 - jj;
    m[0] = (int) (mp1 - 1);    if (mp1 > 13) m[0] -= 12;
    a[0] = (int) (ans - 4715); if (m[0] > 2) a[0] -= 1;
    return tabjours[(int) (jj % 7)];
}

/** Fonction convertissant la date en string à la seconde près
         @param strdate  Sortie : La date formatée A/M/J à XXhXXmnXXs
         @param jdec  Entrée jdec : jour décimal (1.5 = le 1er à midi)
         @param m	  Entrée m : mois (1 à 12)
         @param a	  Entrée a : année (attention 1 avant JC est noté 0, etc..)
         @return      pointeur sur la date
         */
char *dat2str(char *strdate, double jdec, int m, int a)
{
    int j = (int) jdec; // jour entier
    double r = (jdec - j)*24; // heures décimales
    int h = (int) r;
    r = (r - h)*60.; // minutes décimales
    int mn = (int) r;
    int s = (int) (0.5 + (r - mn)*60.); // secondes arrondies
    if (s == 60){
        s = 0; mn = mn + 1;
        if (mn == 60){
            mn = 0; h = h+1;
        }
    }
    sprintf(strdate,"%d/%d/%d à %02dh%02dmn%02ds",a,m,j,h,mn,s);
    return strdate;
}

/** Fonction convertissant la date en string au centième de seconde
         @param strdate Sortie : La date formatée A/M/J à XXhXXmnXXs
         @param jdec  Entrée jdec : jour décimal (1.5 = le 1er à midi)
         @param m	  Entrée m : mois (1 à 12)
         @param a	  Entrée a : année (attention 1 avant JC est noté 0, etc..)
         @return      pointeur sur la date
         */
char *dat2str_2(char *strdate, double jdec, int m, int a)
{
    int j = (int) jdec; // jour entier
    double r = (jdec - j)*24; // heures décimales
    int h = (int) r;
    r = (r - h)*60.; // minutes décimales
    int mn = (int) r;
    double s =  (r - mn)*60.; // secondes decimales
    sprintf(strdate,"%d/%d/%d a %02dh%02dmn%5.2lfs",a,m,j,h,mn,s);
    return strdate;
}

/** Fonction convertissant les jours juliens en string date
         * à la seconde près
         @param strdate Sortie : La date formatée "A/M/J a XXhXXmnXXs"
         @param JJ    Entree : date en jour julien
         @return      pointeur sur la date
         */
char *jul2str (char *strdate, double JJ)
{
    double jdec;
    int a, m;
    jul2dat(&jdec, &m, &a, JJ);
    return dat2str(strdate, jdec, m, a) ;
}

/** Fonction convertissant les jours juliens en string date
         * au centième de seconde
         @param strdate Sortie : La date formatée "A/M/J a XXhXXmnXXs"
         @param JJ    Entree : date en jour julien
         @return      pointeur sur la date
         */
char *jul2str_2 (char *strdate, double JJ)
{
    double jdec;
    int a, m;
    jul2dat(&jdec, &m, &a, JJ);
    return dat2str_2(strdate, jdec, m, a) ;
}

/** Fonction convertissant les jours juliens en tableau date
         * à la seconde près
         @param date  Sortie : int[6] année, mois, jour, hh, mn, s.
         @param JJ    Entree : date en jour julien
         @return      Le jour (Lundi, ...,Dimanche) de la semaine
         */
const char  * jul2tdat(int *date, double JJ)
{
    double jdec;
    int a, m;
    const char  * strj = jul2dat (&jdec, &m, &a, JJ);
    int j = (int) jdec; // jour entier
    double r = (jdec - j)*24; // heures décimales
    int h = (int) r;
    r = (r - h)*60.; // minutes décimales
    int mn = (int) r;
    int s = (int) (0.5 + (r - mn)*60.); // secondes arrondies
    if (s == 60){
        s = 0; mn = mn + 1;
        if (mn == 60){
            mn = 0; h = h+1;
        }
    }
    date[0] = a;
    date[1] = m;
    date[2] = j;
    date[3] = h;
    date[4] = mn;
    date[5] = s;
    return strj ;
}

/** Fonction convertissant un tableau date en jour julien
         @param date  Sortie : int[6] année, mois, jour, hh, mn, s.
         @return     (double) la date en jour julien
         */
double tdat2jul(int *date)
{
    int a = date[0], m = date[1];
    double j = ((date[5]/60. + date[4])/60. + date[3])/24. + date[2];
    return dat2jul(j,m,a);
}

/* Routine de test */
void testjulien()
{

    /** Fonction convertissant un string date du type "2000:05:20 14:00:00" en jour julien
                 @param sdate   Entree : le string date
                 @return     La date en jour julien
         */
    double jj = str19tojul("1998:12:31 23:59:44");
    printf("jj = str19tojul(\"1998:12:31 23:59:44\") = %lf\n", jj);
    /**
         * Calcul ecart deltaET entre temps des ephemerides et temps universel coordonne, approximatif
         * valable uniquement entre 1950 et 2012
         * @param date 	date actuelle en jours julien
         * @return	l'ecart deltaET en jours
         */
    printf("jul2deltaET(jj) = %lf\n", jul2deltaET(jj));

    /** Fonction ajoutant à la date d'un string date du type "2000:05:20 14:00:00"
            un écart specifie en jour decimaux
                 @param sdate   Entree : le string date
                 @param ecart   Entree : l'ecart en jour decimaux
                 @return     La date en jour julien (pointeur sur l'entree)
        */
    printf("add2date(\"2000:05:20 14:00:00\", 5.25) = %s\n",
           add2date("2000:05:20 14:00:00", 5.25));

    /** Fonction convertissant les dates en jour julien
                 @param j    Entree : jour décimal de 1.000 à 31.999 (1.5 : Le 1er à midi)
                 @param m    Entree : mois de 1 à 12
                 @param a    Entree : année (attention 1 avant JC est noté 0, etc..)
                 @return     La date en jour julien
         */
    jj = dat2jul(31.9998148, 12, 1998);
    printf("dat2jul(31.9998148, 12, 1998) = %lf\n", jj);

    /** Fonction donnant le jour de la semaine correspodant à un jour julien donné
                 @param JJ    Entree :  date en jour julien
                 @return      Le jour (Lundi, ...,Dimanche) de la semaine
                 */
    printf("Jour jul2jour(jj) = %s\n", jul2jour(jj));

    /** Fonction convertissant les jours juliens en date decimale
                 @param jdec  Sortie : jdec[0] : jour décimal (1.5 = le 1er à midi)
                 @param m     Sortie : m[0] : mois (1 à 12)
                 @param a     Sortie : a[0] : année (attention 1 avant JC est noté 0, etc..)
                 @param JJ    Entree :  date en jour julien
                 @return      Le jour (Lundi, ...,Dimanche) de la semaine
                 */
    double jdec;
    int mois, annee;
    const char *str = jul2dat(&jdec, &mois, &annee, jj);
    printf("jul2dat(&jdec, &mois, &annee, jj) = %s %lf %d %d\n", str, jdec, mois, annee);

    /** Fonction convertissant la date en string à la seconde près
                 @param strdate  Sortie : La date formatée A/M/J à XXhXXmnXXs
                 @param jdec  Entrée jdec : jour décimal (1.5 = le 1er à midi)
                 @param m	  Entrée m : mois (1 à 12)
                 @param a	  Entrée a : année (attention 1 avant JC est noté 0, etc..)
                 @return      pointeur sur la date
                 */
    char date[128];
    dat2str(date, jdec, mois, annee);
    printf("dat2str(date, jdec, mois, annee) = %s\n", date);

    /** Fonction convertissant la date en string au centième de seconde
                 @param strdate Sortie : La date formatée A/M/J à XXhXXmnXXs
                 @param jdec  Entrée jdec : jour décimal (1.5 = le 1er à midi)
                 @param m     Entrée m : mois (1 à 12)
                 @param a     Entrée a : année (attention 1 avant JC est noté 0, etc..)
                 @return      pointeur sur la date
                 */
    dat2str_2(date, jdec, mois, annee);
    printf("dat2str_2(date, jdec, mois, annee) = %s\n", date);


    /** Fonction convertissant les jours juliens en string date
                 * à la seconde près
                 @param strdate Sortie : La date formatée "A/M/J a XXhXXmnXXs"
                 @param JJ    Entree : date en jour julien
                 @return      pointeur sur la date
                 */
    jul2str (date, jj);
    printf("jul2str (date, jj) = %s\n", date);


    /** Fonction convertissant les jours juliens en string date
                 * au centième de seconde
                 @param strdate Sortie : La date formatée "A/M/J a XXhXXmnXXs"
                 @param JJ    Entree : date en jour julien
                 @return      pointeur sur la date
                 */
    jul2str_2 (date, jj);
    printf("jul2str_2 (date, jj) = %s\n", date);

    /** Fonction convertissant les jours juliens en tableau date
                 * à la seconde près
                 @param date  Sortie : int[6] année, mois, jour, hh, mn, s.
                 @param JJ    Entree : date en jour julien
                 @return      Le jour (Lundi, ...,Dimanche) de la semaine
                 */
    int tdat[6];
    str = jul2tdat(tdat, jj);
    printf("jul2tdat(tdat, jj) = %s %d:%d:%d %dH%dmn%ds\n", str,
           tdat[0], tdat[1], tdat[2], tdat[3], tdat[4], tdat[5]);

    /** Fonction convertissant un tableau date en jour julien
                 @param date  Sortie : int[6] année, mois, jour, hh, mn, s.
                 @return     (double) la date en jour julien
                 */
    printf("tdat2jul(tdat) = %lf\n", tdat2jul(tdat));
    return;
}
